# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Pooling Haiku modules."""

from elegy.module import Module
from typing import Optional, Sequence, Tuple, Union
import warnings

from jax import lax
import jax.numpy as jnp


def _infer_shape(
    x: jnp.ndarray,
    size: Union[int, Sequence[int]],
    channel_axis: Optional[int] = -1,
) -> Tuple[int, ...]:
    """Infer shape for pooling window or strides."""
    if isinstance(size, int):
        if channel_axis and not 0 <= abs(channel_axis) < x.ndim:
            raise ValueError(f"Invalid channel axis {channel_axis} for {x.shape}")
        if channel_axis and channel_axis < 0:
            channel_axis = x.ndim + channel_axis
        return (1,) + tuple(size if d != channel_axis else 1 for d in range(1, x.ndim))
    elif len(size) < x.ndim:
        # Assume additional dimensions are batch dimensions.
        return (1,) * (x.ndim - len(size)) + tuple(size)
    else:
        assert x.ndim == len(size)
        return tuple(size)


_VMAP_SHAPE_INFERENCE_WARNING = (
    "When running under vmap, passing an `int` (except for `1`) for "
    "`window_shape` or `strides` will result in the wrong shape being inferred "
    "because the batch dimension is not visible to Haiku. Please update your "
    "code to specify a full unbatched size. "
    ""
    "For example if you had `pool(x, window_shape=3, strides=1)` before, you "
    "should now pass `pool(x, window_shape=(3, 3, 1), strides=1)`. "
    ""
    "Haiku will assume that any additional dimensions in your input are "
    "batch dimensions, and will pad `window_shape` and `strides` accordingly "
)


def _warn_if_unsafe(window_shape, strides):
    unsafe = lambda size: isinstance(size, int) and size != 1
    if unsafe(window_shape) or unsafe(strides):
        warnings.warn(_VMAP_SHAPE_INFERENCE_WARNING, DeprecationWarning)


def max_pool(
    value: jnp.ndarray,
    window_shape: Union[int, Sequence[int]],
    strides: Union[int, Sequence[int]],
    padding: str,
    channel_axis: Optional[int] = -1,
) -> jnp.ndarray:
    """Max pool.

    Args:
      value: Value to pool.
      window_shape: Shape of the pooling window, an int or same rank as value.
      strides: Strides of the pooling window, an int or same rank as value.
      padding: Padding algorithm. Either ``VALID`` or ``SAME``.
      channel_axis: Axis of the spatial channels for which pooling is skipped,
        used to infer ``window_shape`` or ``strides`` if they are an integer.

    Returns:
      Pooled result. Same rank as value.
    """
    if padding not in ("SAME", "VALID"):
        raise ValueError(f"Invalid padding '{padding}', must be 'SAME' or 'VALID'.")

    _warn_if_unsafe(window_shape, strides)
    window_shape = _infer_shape(value, window_shape, channel_axis)
    strides = _infer_shape(value, strides, channel_axis)

    return lax.reduce_window(value, -jnp.inf, lax.max, window_shape, strides, padding)


def avg_pool(
    value: jnp.ndarray,
    window_shape: Union[int, Sequence[int]],
    strides: Union[int, Sequence[int]],
    padding: str,
    channel_axis: Optional[int] = -1,
) -> jnp.ndarray:
    """Average pool.

    Args:
      value: Value to pool.
      window_shape: Shape of the pooling window, an int or same rank as value.
      strides: Strides of the pooling window, an int or same rank as value.
      padding: Padding algorithm. Either ``VALID`` or ``SAME``.
      channel_axis: Axis of the spatial channels for which pooling is skipped,
        used to infer ``window_shape`` or ``strides`` if they are an integer.

    Returns:
      Pooled result. Same rank as value.

    Raises:
      ValueError: If the padding is not valid.
    """
    if padding not in ("SAME", "VALID"):
        raise ValueError(f"Invalid padding '{padding}', must be 'SAME' or 'VALID'.")

    _warn_if_unsafe(window_shape, strides)
    window_shape = _infer_shape(value, window_shape, channel_axis)
    strides = _infer_shape(value, strides, channel_axis)

    reduce_window_args = (0.0, lax.add, window_shape, strides, padding)
    pooled = lax.reduce_window(value, *reduce_window_args)
    if padding == "VALID":
        # Avoid the extra reduce_window.
        return pooled / jnp.prod(window_shape)
    else:
        # Count the number of valid entries at each input point, then use that for
        # computing average. Assumes that any two arrays of same shape will be
        # padded the same.
        window_counts = lax.reduce_window(jnp.ones_like(value), *reduce_window_args)
        assert pooled.shape == window_counts.shape
        return pooled / window_counts


class MaxPool(Module):
    """Max pool.

    Equivalent to partial application of :func:`max_pool`.
    """

    def __init__(
        self,
        window_shape: Union[int, Sequence[int]],
        strides: Union[int, Sequence[int]],
        padding: str,
        channel_axis: Optional[int] = -1,
        name: Optional[str] = None,
    ):
        """Max pool.

        Args:
          window_shape: Shape of window to pool over. Same rank as value or ``int``.
          strides: Strides for the window. Same rank as value or ``int``.
          padding: Padding algorithm. Either ``VALID`` or ``SAME``.
          channel_axis: Axis of the spatial channels for which pooling is skipped.
        """
        super().__init__(name=name)
        self.window_shape = window_shape
        self.strides = strides
        self.padding = padding
        self.channel_axis = channel_axis

    def call(self, value: jnp.ndarray) -> jnp.ndarray:
        return max_pool(
            value, self.window_shape, self.strides, self.padding, self.channel_axis
        )


class AvgPool(Module):
    """Average pool.

    Equivalent to partial application of :func:`avg_pool`.
    """

    def __init__(
        self,
        window_shape: Union[int, Sequence[int]],
        strides: Union[int, Sequence[int]],
        padding: str,
        channel_axis: Optional[int] = -1,
        name: Optional[str] = None,
    ):
        """Average pool.

        Args:
          window_shape: Shape of window to pool over. Same rank as value or ``int``.
          strides: Strides for the window. Same rank as value or ``int``.
          padding: Padding algorithm. Either ``VALID`` or ``SAME``.
          channel_axis: Axis of the spatial channels for which pooling is skipped.
        """
        super().__init__(name=name)
        self.window_shape = window_shape
        self.strides = strides
        self.padding = padding
        self.channel_axis = channel_axis

    def call(self, value: jnp.ndarray) -> jnp.ndarray:
        return avg_pool(
            value, self.window_shape, self.strides, self.padding, self.channel_axis
        )



import typing as tp
import unittest

import elegy
import jax.numpy as jnp
import numpy as np


class ModelCoreTest(unittest.TestCase):
    def test_init(self):
        N = 0

        class Model(elegy.model.model_core.ModelCore):
            def pred_step(self):
                nonlocal N
                N = N + 1

                return elegy.PredStep.simple(
                    y_pred=None,
                    states=elegy.States(net_params=1, net_states=2),
                )

            def test_step(self):
                _, states, _, _, _ = self.pred_step()
                return elegy.TestStep(0, {}, states.update(metrics_states=3))

            def train_step(self):
                _, logs, states = self.test_step()
                return elegy.TrainStep(logs, states.update(optimizer_states=4))

        model = Model()

        assert N == 0
        assert model.states.net_params != 1
        assert model.states.net_states != 2
        assert model.states.metrics_states != 3
        assert model.states.optimizer_states != 4

        model.maybe_initialize(elegy.types.Mode.pred)

        assert N == 1
        assert model.states.net_params == 1
        assert model.states.net_states == 2

        model.maybe_initialize(elegy.types.Mode.pred)

        assert N == 1

        model.maybe_initialize(elegy.types.Mode.test)

        assert N == 2
        assert model.states.metrics_states == 3

        model.maybe_initialize(elegy.types.Mode.test)

        assert N == 2

        model.maybe_initialize(elegy.types.Mode.train)

        assert N == 3
        assert model.states.optimizer_states == 4

        model.maybe_initialize(elegy.types.Mode.train)

        assert N == 3

    def test_pred_step(self):
        class Model(elegy.model.model_core.ModelCore):
            def pred_step(self, x, net_states, initializing):
                if initializing:
                    states = elegy.States(net_states=0)
                else:
                    states = elegy.States(net_states=net_states + 1)

                return elegy.PredStep.simple(
                    y_pred=1,
                    states=states,
                )

            def test_step(self):
                ...

            def train_step(self):
                ...

        model = Model()
        assert isinstance(model.states.net_states, elegy.types.Uninitialized)

        preds = model.predict_on_batch(x=(np.array(1.0)))
        assert preds == 1
        assert model.states.net_states == 1
        assert isinstance(model.states.net_params, elegy.types.Uninitialized)

        model.run_eagerly = False

        preds = model.predict_on_batch(x=(np.array(1.0)))
        assert preds == 1
        assert model.states.net_states == 2

    def test_test_step(self):
        class Model(elegy.model.model_core.ModelCore):
            def test_step(self, metrics_states, initializing):
                return elegy.TestStep(
                    loss=0.1,
                    logs=dict(loss=1.0),
                    states=elegy.States(metrics_states=0)
                    if initializing
                    else elegy.States(metrics_states=metrics_states + 1),
                )

            def train_step(self):
                ...

        model = Model()
        assert isinstance(model.states.metrics_states, elegy.types.Uninitialized)

        logs = model.test_on_batch(x=(np.array(1.0)), y=(1.0,))
        assert logs["loss"] == 1
        assert model.states.metrics_states == 1
        assert isinstance(model.states.net_params, elegy.types.Uninitialized)

        model.run_eagerly = False

        logs = model.test_on_batch(x=(np.array(1.0)), y=(1.0,))
        assert logs["loss"] == 1
        assert model.states.metrics_states == 2

    def test_train_step(self):
        class Model(elegy.model.model_core.ModelCore):
            def train_step(self, optimizer_states, initializing):
                return elegy.TrainStep(
                    logs=dict(loss=2.0),
                    states=elegy.States(optimizer_states=0)
                    if initializing
                    else elegy.States(optimizer_states=optimizer_states + 1),
                )

        model = Model()
        assert isinstance(model.states.optimizer_states, elegy.types.Uninitialized)

        logs = model.train_on_batch(x=(np.array(1.0)), y=(1.0,))
        assert logs["loss"] == 2
        assert model.states.optimizer_states == 1
        assert isinstance(model.states.net_params, elegy.types.Uninitialized)

        model.run_eagerly = False

        logs = model.train_on_batch(x=(np.array(1.0)), y=(1.0,))
        assert logs["loss"] == 2
        assert model.states.optimizer_states == 2



import typing as tp

from elegy import hooks, types, utils
from elegy.module import Module

from .generalized_module import GeneralizedModule, register_module_for


@register_module_for(Module)
class ElegyModule(GeneralizedModule):
    def __init__(self, module: Module):
        self.module = module

    def init(self, rng: types.RNGSeq) -> tp.Callable[..., types.OutputStates]:
        def _lambda(*args, **kwargs):

            y_pred, collections = utils.inject_dependencies(self.module.init(rng=rng))(
                *args,
                **kwargs,
            )
            assert isinstance(collections, dict)

            net_params = collections.pop("parameters", {})
            net_states = collections

            return types.OutputStates(y_pred, net_params, net_states)

        return _lambda

    def update(
        self,
        params: tp.Optional[types.ModuleParams],
        states: tp.Optional[types.ModuleStates],
    ) -> tp.Tuple[tp.Optional[types.ModuleParams], tp.Optional[types.ModuleStates]]:

        collections = states

        if not utils.none_or_uninitialized(params):
            collections = collections.copy() if collections is not None else {}
            collections["parameters"] = params

        if not utils.none_or_uninitialized(collections):
            assert collections is not None
            self.module.set_default_parameters(collections)

        return params, states

    def apply(
        self,
        params: tp.Any,
        states: tp.Any,
        training: bool,
        rng: types.RNGSeq,
    ) -> tp.Callable[..., types.OutputStates]:
        def _lambda(*args, **kwargs):
            collections = states.copy() if states is not None else {}
            if params is not None:
                collections["parameters"] = params

            y_pred, collections = utils.inject_dependencies(
                self.module.apply(collections, training=training, rng=rng),
            )(
                *args,
                **kwargs,
            )
            assert isinstance(collections, dict)

            net_params = collections.pop("parameters", {})
            net_states = collections

            return types.OutputStates(y_pred, net_params, net_states)

        return _lambda

    def get_summary_params(
        self,
        path: types.Path,
        module: tp.Any,
        value: tp.Any,
        include_submodules: bool,
        net_params: types.NetParams,
        net_states: types.NetStates,
    ) -> tp.Tuple[tp.Optional[types.Pytree], tp.Optional[types.Pytree]]:

        if net_params is None:
            params_tree = None
        else:
            params_tree = utils.get_path_params(path, net_params)
            # filter only params
            if not include_submodules and params_tree is not None:
                assert isinstance(module, Module)
                params_tree = {
                    name: value
                    for name, value in params_tree.items()
                    if name in module._spec
                }

        if net_states is None:
            states_tree = None
        else:
            states_tree = {
                collection: utils.get_path_params(path, states)
                for collection, states in net_states.items()
            }
            # filter only params
            if not include_submodules:
                states_tree = {
                    collection: {
                        name: value
                        for name, value in states.items()
                        if name in module._spec
                    }
                    for collection, states in states_tree.items()
                    if states is not None
                }

        return params_tree, states_tree



from unittest import TestCase

import jax

import elegy
from elegy.generalized_module.generalized_module import generalize
import jax.numpy as jnp


class ElegyModuleTest(TestCase):
    def test_basic(self):
        class M(elegy.Module):
            def call(self, x):
                n = self.add_parameter("n", lambda: 0, trainable=False)
                w = self.add_parameter("w", lambda: 2.0)

                self.update_parameter("n", n + 1)

                key = self.next_key()

                return x * w

        gm = generalize(M())
        rng = elegy.RNGSeq(42)

        y_true, params, states = gm.init(rng)(x=3.0, y=1)

        assert y_true == 6
        assert params["w"] == 2
        assert states["states"]["n"] == 0

        params["w"] = 10.0
        y_true, params, states = gm.apply(params, states, training=True, rng=rng)(
            x=3.0, y=1
        )

        assert y_true == 30
        assert params["w"] == 10
        assert states["states"]["n"] == 1

    def test_summaries(self):
        class ModuleC(elegy.Module):
            def call(self, x):
                c1 = self.add_parameter("c1", lambda: jnp.ones([5]))
                c2 = self.add_parameter("c2", lambda: jnp.ones([6]), trainable=False)

                x = jax.nn.relu(x)
                self.add_summary("relu", jax.nn.relu, x)

                return x

        class ModuleB(elegy.Module):
            def call(self, x):
                b1 = self.add_parameter("b1", lambda: jnp.ones([3]))
                b2 = self.add_parameter("b2", lambda: jnp.ones([4]), trainable=False)

                x = ModuleC()(x)

                x = jax.nn.relu(x)
                self.add_summary("relu", jax.nn.relu, x)

                return x

        class ModuleA(elegy.Module):
            def call(self, x):
                a1 = self.add_parameter("a1", lambda: jnp.ones([1]))
                a2 = self.add_parameter("a2", lambda: jnp.ones([2]), trainable=False)

                x = ModuleB()(x)

                x = jax.nn.relu(x)
                self.add_summary("relu", jax.nn.relu, x)

                return x

        model = elegy.Model(ModuleA())

        summary_text = model.summary(x=jnp.ones([10, 2]), depth=1, return_repr=True)
        assert summary_text is not None

        lines = summary_text.split("\n")

        assert "module_b" in lines[7]
        assert "ModuleB" in lines[7]
        assert "(10, 2)" in lines[7]
        assert "8" in lines[7]
        assert "32 B" in lines[7]
        assert "10" in lines[7]
        assert "40 B" in lines[7]

        assert "relu" in lines[9]
        assert "(10, 2)" in lines[9]

        assert "*" in lines[11]
        assert "ModuleA" in lines[11]
        assert "(10, 2)" in lines[11]
        assert "1" in lines[11]
        assert "4 B" in lines[11]
        assert "2" in lines[11]
        assert "8 B" in lines[11]

        assert "21" in lines[13]
        assert "84 B" in lines[13]

        assert "9" in lines[14]
        assert "36 B" in lines[14]

        assert "12" in lines[15]
        assert "48 B" in lines[15]



from unittest import TestCase
import elegy

import tensorflow.keras as tfk
import numpy as np
import jax.numpy as jnp


class MeanAbsolutePercentageErrorTest(TestCase):
    #
    def test_basic(self):

        y_true = np.random.random(size=(5, 6, 7))
        y_pred = np.random.random(size=(5, 6, 7))

        assert np.allclose(
            tfk.metrics.MeanAbsolutePercentageError()(y_true, y_pred),
            elegy.metrics.MeanAbsolutePercentageError().call_with_defaults()(
                jnp.asarray(y_true), jnp.asarray(y_pred)
            ),
        )

    #
    def test_cumulative(self):

        tm = tfk.metrics.MeanAbsolutePercentageError()
        em = elegy.metrics.MeanAbsolutePercentageError()

        # 1st run
        y_true = np.random.random(size=(5, 6, 7))
        y_pred = np.random.random(size=(5, 6, 7))

        assert np.allclose(
            tm(y_true, y_pred),
            em.call_with_defaults()(jnp.asarray(y_true), jnp.asarray(y_pred)),
        )

        # 2nd run
        y_true = np.random.random(size=(5, 6, 7))
        y_pred = np.random.random(size=(5, 6, 7))

        assert np.allclose(
            tm(y_true, y_pred),
            em.call_with_defaults()(jnp.asarray(y_true), jnp.asarray(y_pred)),
        )



from .accuracy import Accuracy, accuracy
from .precision import Precision, precision
from .recall import Recall, recall
from .f1 import F1, f1
from .binary_accuracy import BinaryAccuracy, binary_accuracy
from .binary_crossentropy import BinaryCrossentropy, binary_crossentropy
from .categorical_accuracy import CategoricalAccuracy, categorical_accuracy
from .mean import Mean
from .mean_absolute_error import MeanAbsoluteError, mean_absolute_error
from .mean_squared_error import MeanSquaredError, mean_squared_error
from .mean_absolute_percentage_error import (
    MeanAbsolutePercentageError,
    mean_absolute_percentage_error,
)
from .metric import Metric
from .reduce import Reduce, reduce, Reduction
from .sparse_categorical_accuracy import (
    SparseCategoricalAccuracy,
    sparse_categorical_accuracy,
)
from .sum import Sum

__all__ = [
    "Accuracy",
    "BinaryAccuracy",
    "BinaryCrossentropy",
    "CategoricalAccuracy",
    "F1",
    "Mean",
    "MeanAbsoluteError",
    "MeanAbsolutePercentageError",
    "MeanSquaredError",
    "Metric",
    "Precision",
    "Recall",
    "Reduce",
    "Reduction",
    "SparseCategoricalAccuracy",
    "Sum",
    "accuracy",
    "binary_accuracy",
    "binary_crossentropy",
    "categorical_accuracy",
    "f1",
    "mean_absolute_error",
    "mean_absolute_percentage_error",
    "mean_squared_error",
    "precision",
    "recall",
    "reduce",
    "sparse_categorical_accuracy",
]



from elegy import types
from unittest import TestCase

import jax
import jax.numpy as jnp
import numpy as np
import pytest

import elegy


class NewModuleTest(TestCase):
    def test_set_parameters(self):
        class M(elegy.Module):
            def call(self, x):
                self.add_parameter("m", lambda: 1)
                self.add_parameter("n", lambda: 2)
                return x

        m = M()
        m.init()(2.0)

        collections = {"parameters": {"m": 1, "n": 10}}
        m.set_default_parameters(collections)

        assert m.m == 1
        assert m.n == 10

        with pytest.raises(ValueError):
            collections = {
                "parameters": {"m": 100},
                "xyz": {"n": 100},
            }
            m.set_default_parameters(collections)

        assert m.m == 1
        assert m.n == 10

    def test_get_parameters(self):
        class M(elegy.Module):
            def call(self, x):
                self.add_parameter("m", lambda: 1)
                self.add_parameter("n", lambda: 2)
                return x

        m = M()
        m.init()(2.0)

        collections = {"parameters": {"m": 1, "n": 10}}

        m.set_default_parameters(collections)

        collections = m.get_default_parameters()

        assert collections == {"parameters": {"m": 1, "n": 10}}

    def test_basic(self):
        class M(elegy.Module):
            def call(self, x):
                n = self.add_parameter("n", lambda: 1)
                return n + x

        m = M()

        y, collections = m.init()(2.0)

        assert collections == {"parameters": {"n": 1}}
        assert y == 3

    def test_basic_call_error(self):
        class M(elegy.Module):
            def call(self, x):
                n = self.add_parameter("n", lambda: 1)
                return n + x

        m = M()

        with pytest.raises(types.NoContext):
            y = m(2.0)

    def test_basic_call(self):
        class M(elegy.Module):
            def call(self, x):
                n = self.add_parameter("n", lambda: 1)
                return n + x

        m = M()

        y = m.call_with_defaults()(2.0)
        collections = m.get_default_parameters()

        assert y == 3
        assert collections == {"parameters": {"n": 1}}

        # update params
        m.set_default_parameters({"parameters": {"n": 20}})
        y = m.call_with_defaults()(2.0)
        collections = m.get_default_parameters()

        assert y == 22
        assert collections == {"parameters": {"n": 20}}

    def test_basic_apply(self):
        class M(elegy.Module):
            def call(self, x):
                n = self.add_parameter("n", lambda: 1)
                return n + x

        m = M()

        with pytest.raises(ValueError):
            m.get_default_parameters()

        y, collections = m.init()(2.0)

        assert y == 3
        assert collections == {"parameters": {"n": 1}}

        # set default parameters
        m.set_default_parameters(collections)

        # run with new params
        y, collections = m.apply({"parameters": {"n": 5}})(2.0)

        assert y == 7
        assert collections == {"parameters": {"n": 5}}

        # check defaults are not modify by apply.
        default_params = m.get_default_parameters()
        assert default_params == {"parameters": {"n": 1}}

    def test_basic_compose(self):
        class A(elegy.Module):
            def call(self, x):
                b = self.add_parameter("b", lambda: 10)
                return b + x

        class M(elegy.Module):
            def call(self, x):
                n = self.add_parameter("n", lambda: 1)
                x = A()(x)
                return n + x

        m = M()

        y, collections = m.init()(2.0)

        assert y == 13
        assert collections == {"parameters": {"n": 1, "a": {"b": 10}}}

    def test_basic_list_submodules(self):
        class A(elegy.Module):
            def call(self, x):
                b = self.add_parameter("b", lambda: 10)
                return b + x

        class M(elegy.Module):
            def __init__(self):
                super().__init__()
                self.ais = [A(), A()]

            def call(self, x):
                n = self.add_parameter("n", lambda: 1)
                x = self.ais[1](x)
                return n + x

        m = M()

        with elegy.hooks.context(set_all=True):
            y, collections = m.init()(2.0)
            summaries = elegy.hooks.get_summaries()

        assert summaries == [
            (
                ("ais/1",),
                m.ais[1],
                12,
            ),
            (
                (),
                m,
                13,
            ),
        ]
        assert collections == {
            "parameters": {
                "n": 1,
                "ais/0": {},
                "ais/1": {"b": 10},
            },
        }
        assert y == 13

    def test_basic_dynamic_submodules(self):
        class A(elegy.Module):
            def call(self, x):
                b = self.add_parameter("b", lambda: 10)
                return b + x

        class M(elegy.Module):
            a: A
            a_1: A

            def call(self, x):
                ais = [A(), A()]
                n = self.add_parameter("n", lambda: 1)
                x = ais[1](x)
                return n + x

        m = M()

        with elegy.hooks.context(set_all=True):
            y, params = m.init()(2.0)
            summaries = elegy.hooks.get_summaries()

        assert summaries == [
            (
                ("a_1",),
                m.a_1,
                12,
            ),
            (
                (),
                m,
                13,
            ),
        ]
        assert params == {
            "parameters": {
                "n": 1,
                "a": {},
                "a_1": {
                    "b": 10,
                },
            },
        }
        assert y == 13

    def test_basic_update(self):
        class M(elegy.Module):
            def call(self, x):
                n = self.add_parameter("n", lambda: 1)
                self.update_parameter("n", n + 1)
                return n + x

        m = M()

        y, collections = m.init()(2.0)

        assert y == 3
        assert collections == {"parameters": {"n": 1}}

        y, collections = m.apply(collections)(2.0)
        assert y == 3

        y, collections = m.apply(collections)(2.0)
        assert y == 4

        # jit has to use functional API
        y, collections = m.apply_jit(collections)(2.0)
        assert y == 5

        # error is raised if no parameters are given
        with pytest.raises(TypeError):
            y, collections = m.apply_jit()(2.0)


class ModuleTest(TestCase):
    class Linear(elegy.Module):
        def __init__(self, units):
            super().__init__()
            self.units = units

        def call(self, x):
            w = self.add_parameter("w", lambda: jnp.ones([x.shape[-1], self.units]))
            b = self.add_parameter("b", lambda: jnp.ones([self.units]))
            n = self.add_parameter("n", lambda: jnp.zeros([]), trainable=False)

            self.update_parameter("n", n + 1)

            y = jnp.dot(x, w) + b

            path = elegy.module.get_module_path_str(self)
            elegy.hooks.add_loss(f"activation_sum", jnp.sum(y))
            elegy.hooks.add_metric(f"{path}/activation_mean", jnp.mean(y))

            return y

    class MyModule(elegy.Module):
        def __init__(self):
            super().__init__()
            self.linear = ModuleTest.Linear(6)
            self.linear1 = ModuleTest.Linear(7)

        def call(self, x) -> np.ndarray:
            x = self.linear(x)
            x = self.linear1(x)
            self.bias = self.add_parameter(
                "bias", lambda: jnp.ones([x.shape[-1]], jnp.float32)
            )
            return x + self.bias * 10

    def test_basic(self):
        x = np.random.uniform(-1, 1, size=(4, 5))
        m = ModuleTest.MyModule()
        y = m.call_with_defaults()(x)
        assert y.shape == (4, 7)

    def test_get_parameters(self):
        x = np.random.uniform(-1, 1, size=(4, 5))

        m = ModuleTest.MyModule()

        y, collections = m.init()(x)

        parameters, states = (
            collections["parameters"],
            collections["states"],
        )

        assert "bias" in parameters
        assert "linear" in parameters
        assert "w" in parameters["linear"]
        assert "b" in parameters["linear"]
        assert states["linear"]["n"] == 0
        assert states["linear1"]["n"] == 0
        assert "linear1" in parameters

        with elegy.hooks.context(set_all=True):
            y, collections = m.apply(collections)(x)
            # y2: jnp.ndarray = m.call_jit(x)

            losses = elegy.hooks.get_losses()
            metrics = elegy.hooks.get_metrics()
            summaries = elegy.hooks.get_summaries()

        assert losses
        assert metrics
        assert summaries

        parameters, states = (
            collections["parameters"],
            collections["states"],
        )

        assert y.shape == (4, 7)
        assert "bias" in parameters
        assert "linear" in parameters
        assert "w" in parameters["linear"]
        assert "b" in parameters["linear"]
        assert states["linear"]["n"] == 1
        assert "linear1" in parameters

        assert "activation_sum_loss" in losses
        assert "linear/activation_mean" in metrics
        assert "linear1/activation_mean" in metrics

        assert summaries[0][:2] == (("linear",), m.linear)
        assert summaries[0][2].shape == (4, 6)
        assert summaries[1][:2] == (("linear1",), m.linear1)
        assert summaries[1][2].shape == (4, 7)
        assert summaries[2][:2] == ((), m)
        assert summaries[2][2].shape == (4, 7)

        m.set_default_parameters(jax.tree_map(lambda x: -x, collections))

        collections = m.get_default_parameters()
        parameters, states = (
            collections["parameters"],
            collections["states"],
        )

        assert parameters["bias"][0] == -1
        assert m.linear.get_default_parameters()["parameters"]["w"][0, 0] == -1
        assert m.linear.get_default_parameters()["parameters"]["b"][0] == -1
        assert m.linear1.get_default_parameters()["parameters"]["w"][0, 0] == -1
        assert m.linear1.get_default_parameters()["parameters"]["b"][0] == -1

        current_collection = m.get_default_parameters()

        m.clear_default_parameters()

        collections = m.get_default_parameters()

        assert jax.tree_leaves(collections) == []
        assert elegy.utils.parameters_count(collections) == 0

        m.set_default_parameters(current_collection)

        assert m.get_default_parameters()["parameters"]["bias"][0] == -1
        assert m.linear.get_default_parameters()["parameters"]["w"][0, 0] == -1
        assert m.linear.get_default_parameters()["parameters"]["b"][0] == -1
        assert m.linear1.get_default_parameters()["parameters"]["w"][0, 0] == -1
        assert m.linear1.get_default_parameters()["parameters"]["b"][0] == -1

    def test_set_parameters_shape_check(self):
        x = np.random.uniform(-1, 1, size=(4, 5))
        m = ModuleTest.MyModule()

        m.init(set_defaults=True)(x)

        params0 = m.get_default_parameters()
        # new random params
        params1 = jax.tree_map(lambda x: np.random.random(x.shape), params0)
        # set a parameter with incorrect shape
        params1["parameters"]["linear1"]["w"] = np.zeros([10, 10])

        # should raise error when trying to set the incorrect params
        with pytest.raises(ValueError):
            m.set_default_parameters(params1)

        # the parameters should not have changed despite the error
        assert np.all(
            jax.tree_leaves(
                jax.tree_multimap(
                    lambda x, y: np.allclose(x, y), params0, m.get_default_parameters()
                )
            )
        )

        # should not raise an error
        with pytest.raises(ValueError):
            m.set_default_parameters(params1)

        # linear1.w should not change
        assert np.allclose(
            m.get_default_parameters()["parameters"]["linear1"]["w"],
            params0["parameters"]["linear1"]["w"],
        )
        # but all others
        assert np.all(
            jax.tree_multimap(
                lambda x, y: np.allclose(x, y) if x.shape == y.shape else True,
                params1,
                m.get_default_parameters(),
            )
        )

        # remove a parameter
        params1["parameters"]["linear1"].pop("w")
        # should raise with check_missing=True
        with pytest.raises(ValueError):
            m.set_default_parameters(params1)


class ModuleDynamicTest(TestCase):
    class Linear(elegy.Module):
        w: jnp.ndarray
        b: jnp.ndarray

        def __init__(self, units):
            super().__init__()
            self.units = units

        def call(self, x):
            w = self.add_parameter("w", lambda: jnp.ones([x.shape[-1], self.units]))
            b = self.add_parameter("b", lambda: jnp.ones([self.units]))
            n = self.add_parameter("n", lambda: jnp.zeros([]), trainable=False)

            self.update_parameter("n", n + 1)

            y = jnp.dot(x, w) + b

            path = elegy.module.get_module_path_str(self)
            elegy.hooks.add_loss(f"activation_sum", jnp.sum(y))
            elegy.hooks.add_metric(f"{path}/activation_mean", jnp.mean(y))

            return y

    class MyModule(elegy.Module):
        linear: "ModuleDynamicTest.Linear"
        linear_1: "ModuleDynamicTest.Linear"

        def call(self, x) -> np.ndarray:
            x = ModuleDynamicTest.Linear(6)(x)
            x = ModuleDynamicTest.Linear(7)(x)
            self.bias = self.add_parameter("bias", lambda: jnp.ones([x.shape[-1]]))
            return x + self.bias * 10

    def test_basic(self):
        x = np.random.uniform(-1, 1, size=(4, 5))
        m = ModuleDynamicTest.MyModule()
        y: jnp.ndarray = m.call_with_defaults()(x)
        assert y.shape == (4, 7)
        print(m.get_default_parameters)

    def test_get_parameters(self):
        x = np.random.uniform(-1, 1, size=(4, 5))
        m = ModuleDynamicTest.MyModule()

        m.init_jit(set_defaults=True)(x)

        collections = m.get_default_parameters()
        parameters, states = (
            collections["parameters"],
            collections["states"],
        )

        assert "bias" in parameters
        assert "linear" in parameters
        assert "w" in parameters["linear"]
        assert "b" in parameters["linear"]
        assert m.linear.get_default_parameters()["states"]["n"] == 0
        assert states["linear"]["n"] == 0
        assert "linear_1" in parameters

        with elegy.hooks.context(set_all=True):
            # y: jnp.ndarray = m.call_with_defaults()(x)
            collections = m.get_default_parameters()
            y: jnp.ndarray
            y, collections = m.apply_jit(collections)(x)
            m.set_default_parameters(collections)
            parameters, states = (
                collections["parameters"],
                collections["states"],
            )

            losses = elegy.hooks.get_losses()
            metrics = elegy.hooks.get_metrics()
            summaries = elegy.hooks.get_summaries()

        assert losses is not None
        assert metrics
        assert summaries

        assert y.shape == (4, 7)
        assert "bias" in parameters
        assert "linear" in parameters
        assert "w" in parameters["linear"]
        assert "b" in parameters["linear"]
        assert m.linear.get_default_parameters()["states"]["n"] == 1
        assert states["linear"]["n"] == 1
        assert "linear_1" in parameters

        assert "activation_sum_loss" in losses
        assert "linear/activation_mean" in metrics
        assert "linear_1/activation_mean" in metrics

        assert summaries[0][:2] == (("linear",), m.linear)
        assert summaries[0][2].shape == (4, 6)
        assert summaries[1][:2] == (("linear_1",), m.linear_1)
        assert summaries[1][2].shape == (4, 7)
        assert summaries[2][:2] == ((), m)
        assert summaries[2][2].shape == (4, 7)

        m.set_default_parameters(jax.tree_map(lambda x: -x, m.get_default_parameters()))

        collections = m.get_default_parameters()
        parameters, states = (
            collections["parameters"],
            collections["states"],
        )

        assert parameters["bias"][0] == -1
        assert m.linear.get_default_parameters()["parameters"]["w"][0, 0] == -1
        assert m.linear.get_default_parameters()["parameters"]["b"][0] == -1
        assert m.linear_1.get_default_parameters()["parameters"]["w"][0, 0] == -1
        assert m.linear_1.get_default_parameters()["parameters"]["b"][0] == -1

        current_parameters = m.get_default_parameters()

        m.clear_default_parameters()

        assert jax.tree_leaves(m.get_default_parameters()) == []
        assert elegy.utils.parameters_count(m.get_default_parameters()) == 0

        m.set_default_parameters(current_parameters)

        assert m.get_default_parameters()["parameters"]["bias"][0] == -1
        assert m.linear.get_default_parameters()["parameters"]["w"][0, 0] == -1
        assert m.linear.get_default_parameters()["parameters"]["b"][0] == -1
        assert m.linear_1.get_default_parameters()["parameters"]["w"][0, 0] == -1
        assert m.linear_1.get_default_parameters()["parameters"]["b"][0] == -1

    def test_auto_init(self):
        x = np.random.uniform(-1, 1, size=(4, 5))
        m = ModuleDynamicTest.MyModule()

        m.call_with_defaults()(x)

        # THESE:
        assert m.linear.get_default_parameters()["states"]["n"] == 1
        assert m.get_default_parameters()["states"]["linear"]["n"] == 1

        assert "bias" in m.get_default_parameters()["parameters"]
        assert "linear" in m.get_default_parameters()["parameters"]
        assert "w" in m.get_default_parameters()["parameters"]["linear"]
        assert "b" in m.get_default_parameters()["parameters"]["linear"]
        assert "linear_1" in m.get_default_parameters()["parameters"]


class TestTransforms(TestCase):
    def test_simple_0(self):

        total_called = 0

        @jax.jit
        def f(params, x):
            nonlocal total_called

            total_called += 1

            y = params["w"] * x + params["b"]
            params["w"] += 1.0

            return params, y

        params = {"w": jnp.array(1.0), "b": jnp.array(2.0)}
        params, outputs = f(params, 1)
        # m.set_default_parameters(params)

        print(params, type(params))

        assert total_called == 1

        params = {"b": params["b"], "w": params["w"]}
        params, outputs = f(params, 1)
        # m.set_default_parameters(params)

        assert total_called == 1

    def test_simple_defaults(self):

        total_called = 0

        class SomeModule(elegy.Module):
            n: jnp.ndarray

            def call(self, x):
                nonlocal total_called

                total_called += 1

                n = self.add_parameter("n", lambda: jnp.array(0))
                self.update_parameter("n", n + 1)

                if self.is_training():
                    return x + 1
                else:
                    return x - 1

        m = SomeModule()

        @jax.jit
        def f(params, x):

            m.set_default_parameters(params)

            outputs = m.call_with_defaults()(x)

            return m.get_default_parameters(), outputs

        assert total_called == 0

        _, collections = m.init(set_defaults=True)(1)
        assert total_called == 1

        collections, outputs = f(collections, 1)
        m.set_default_parameters(collections)

        assert total_called == 2

        collections = m.get_default_parameters()
        print(collections)
        collections, outputs = f(collections, 1)
        m.set_default_parameters(collections)

        assert total_called == 2

    def test_simple_apply(self):

        total_called = 0

        class SomeModule(elegy.Module):
            n: jnp.ndarray

            def call(self, x):
                nonlocal total_called

                total_called += 1

                n = self.add_parameter("n", lambda: jnp.array(0))
                self.update_parameter("n", n + 1)

                if self.is_training():
                    return x + 1
                else:
                    return x - 1

        m = SomeModule()

        @jax.jit
        def f(collections, x):

            y, collections = m.apply(collections)(x)

            return collections, y

        assert total_called == 0

        y, collections = m.init()(1)

        assert total_called == 1

        collections, y = f(collections, 1)

        assert total_called == 2

        collections, y = f(collections, 1)

        assert total_called == 2

    def test_jit(self):
        total_called = 0

        class SomeModule(elegy.Module):
            n: jnp.ndarray

            def call(self, x):
                nonlocal total_called

                total_called += 1

                n = self.add_parameter("n", lambda: jnp.array(0))
                self.update_parameter("n", n + 1)

                if elegy.hooks.losses_active():
                    return x + 1
                else:
                    return x - 1

        m = SomeModule()
        assert total_called == 0

        with elegy.hooks.context(losses=True):
            m.init_jit(set_defaults=True)(0)
            # triggers call because its the first time
            assert total_called == 1
            assert m.n == 0

        with elegy.hooks.context(losses=True):
            y = m.call_with_defaults_jit()(0)

            assert y == 1
            assert m.n == 1
            assert total_called == 2

        with elegy.hooks.context(losses=True):
            y = m.call_with_defaults_jit()(0)
            assert m.n == 2
            assert total_called == 2

        with elegy.hooks.context(losses=False):
            y = m.call_with_defaults_jit()(0)
            assert y == -1
            # triggers call because training changed and is static
            assert total_called == 3
            assert m.n == 3

        with elegy.hooks.context(losses=False):
            y = m.call_with_defaults_jit()(0)
            assert y == -1
            # does not trigger call function for training = True exists
            assert total_called == 3
            assert m.n == 4

        with elegy.hooks.context(losses=True):
            y = m.call_with_defaults_jit()(0)
            assert y == 1
            # does not trigger call function for training = True exists
            assert total_called == 3
            assert m.n == 5

        with elegy.hooks.context(losses=True, summaries=True):
            y = m.call_with_defaults_jit()(0)
            assert y == 1
            # triggers call because summaries are now present
            assert total_called == 4
            assert m.n == 6

        with elegy.hooks.context(losses=False, summaries=True):
            y = m.call_with_defaults_jit()(0)
            assert y == -1
            # triggers call because configuration training=False,  is new
            assert total_called == 5
            assert m.n == 7

    def test_jit_auto_init(self):
        total_called = 0

        class SomeModule(elegy.Module):
            n: jnp.ndarray

            def call(self, x):
                nonlocal total_called

                total_called += 1

                n = self.add_parameter("n", lambda: jnp.array(0))
                self.update_parameter("n", n + 1)

                if elegy.hooks.losses_active():
                    return x + 1
                else:
                    return x - 1

        m = SomeModule()

        assert total_called == 0

        with elegy.hooks.context(losses=True):
            m.call_with_defaults_jit()(0)
            assert total_called == 2

            assert m.n == 1

            y = m.call_with_defaults_jit()(0)

            assert y == 1
            assert m.n == 2
            assert total_called == 2

            y = m.call_with_defaults_jit()(0)
            assert m.n == 3
            assert total_called == 2


class TestOthers(TestCase):
    def test_module_system_docs(self):
        class Linear(elegy.Module):
            def __init__(self, n_out):
                super().__init__()
                self.n_out = n_out

            def call(self, x):
                w = self.add_parameter(
                    "w",
                    lambda: elegy.initializers.RandomUniform()(
                        shape=[x.shape[-1], self.n_out],
                        dtype=jnp.float32,
                    ),
                )
                b = self.add_parameter("b", lambda: jnp.zeros(shape=[self.n_out]))

                return jnp.dot(x, w) + b

        class MLP(elegy.Module):
            def call(self, x):
                x = Linear(64)(x)
                x = jax.nn.relu(x)
                x = Linear(32)(x)
                x = jax.nn.relu(x)
                x = Linear(1)(x)
                return x

        def loss_fn(parameters, x, y):
            y_pred, _ = mlp.apply(dict(parameters=parameters))(x)
            return jnp.mean(jnp.square(y - y_pred))

        def update(parameters, x, y):
            loss, gradients = jax.value_and_grad(loss_fn)(parameters, x, y)
            parameters = jax.tree_multimap(
                lambda p, g: p - 0.01 * g, parameters, gradients
            )

            return loss, parameters

        x = np.random.uniform(size=(15, 3))
        y = np.random.uniform(size=(15, 1))
        mlp = MLP()

        y_pred, collections = mlp.init(rng=elegy.RNGSeq(42))(x)

        parameters = collections["parameters"]

        update_jit = jax.jit(update)

        for step in range(1):
            loss, parameters = update_jit(parameters, x, y)

        mlp.set_default_parameters(dict(parameters=parameters))



from .batch_normalization import BatchNormalization
from .conv import Conv1D, Conv2D, Conv3D, ConvND
from .dropout import Dropout
from .flatten import Flatten, Reshape
from .linear import Linear
from .sequential_module import Sequential, sequential

from .layer_normalization import LayerNormalization, InstanceNormalization
from .embedding import Embedding, EmbedLookupStyle
from .pool import MaxPool, AvgPool
from .moving_averages import EMAParamsTree

__all__ = [
    "EMAParamsTree",
    "BatchNormalization",
    "Conv1D",
    "Conv2D",
    "Conv3D",
    "ConvND",
    "Dropout",
    "Flatten",
    "Reshape",
    "Linear",
    "Sequential",
    "sequential",
    "LayerNormalization",
    "InstanceNormalization",
    "Embedding",
    "EmbedLookupStyle",
    "MaxPool",
    "AvgPool",
]



import elegy
import jax.numpy as jnp
import numpy as np
import unittest


class TestModelBase(unittest.TestCase):
    def test_predict(self):
        class Model(elegy.model.model_base.ModelBase):
            def pred_step(self, x, net_states, initializing):
                if initializing:
                    states = elegy.States(net_states=0)
                else:
                    states = elegy.States(net_states=net_states + 1)

                return elegy.PredStep.simple(x + 1.0, states)

        model = Model()

        x = np.random.uniform(size=(100, 1))
        y = model.predict(x, batch_size=50)

        assert np.allclose(y, x + 1.0)
        assert model.states.net_states == 2
        assert model.initial_states.net_states == 0

    def test_evaluate(self):
        class Model(elegy.model.model_base.ModelBase):
            def test_step(self, x, metrics_states, initializing):
                if initializing:
                    states = elegy.States(metrics_states=0)
                else:
                    states = elegy.States(metrics_states=metrics_states + 1)

                return elegy.TestStep(
                    loss=0.1,
                    logs=dict(loss=jnp.sum(x)),
                    states=states,
                )

            def train_step(self):
                ...

        model = Model(run_eagerly=True)

        x = np.random.uniform(size=(100, 1))

        logs = model.evaluate(x, batch_size=100)
        assert np.allclose(logs["loss"], np.sum(x))
        assert model.states.metrics_states == 1

        logs = model.evaluate(x, batch_size=50)
        assert np.allclose(logs["loss"], np.sum(x[50:]))
        assert model.states.metrics_states == 2

    def test_fit(self):
        class Model(elegy.model.model_base.ModelBase):
            def train_step(self, x, optimizer_states, initializing):
                if initializing:
                    states = elegy.States(optimizer_states=0)
                else:
                    states = elegy.States(optimizer_states=optimizer_states + 1)

                return elegy.TrainStep(
                    logs=dict(loss=jnp.sum(x)),
                    states=states,
                )

        model = Model()

        x = np.random.uniform(size=(100, 1))

        history = model.fit(x, batch_size=100)
        assert np.allclose(history.history["loss"], np.sum(x))
        assert model.states.optimizer_states == 1

        history = model.fit(x, batch_size=50, shuffle=False)
        assert np.allclose(history.history["loss"], np.sum(x[50:]))
        assert model.states.optimizer_states == 3



import elegy
from elegy.generalized_optimizer.generalized_optimizer import generalize_optimizer
import optax


def test_basic():

    w = 2.0
    grads = 1.5
    lr = 1.0
    rng = elegy.RNGSeq(42)

    go = generalize_optimizer(optax.sgd(lr))

    states = go.init(rng, w)
    w, states = go.apply(w, grads, states, rng)

    assert w == 0.5



# Implementation based on Tensorflow Keras and Haiku
# Tensorflow: https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/keras/layers/normalization.py#L46
# Haiku: https://github.com/deepmind/dm-haiku/blob/master/haiku/_src/batch_norm.py#L39#L194

import typing as tp
from typing import Optional, Sequence

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from haiku._src import utils as haiku_utils

from elegy import initializers, module, types
from elegy import module
from elegy import hooks
from elegy.nn.moving_averages import ExponentialMovingAverage


class BatchNormalization(module.Module):
    """Normalizes inputs to maintain a mean of ~0 and stddev of ~1.

    See: https://arxiv.org/abs/1502.03167.

    There are many different variations for how users want to manage scale and
    offset if they require them at all. These are:

      - No scale/offset in which case ``create_*`` should be set to ``False`` and
        ``scale``/``offset`` aren't passed when the module is called.
      - Trainable scale/offset in which case ``create_*`` should be set to
        ``True`` and again ``scale``/``offset`` aren't passed when the module is
        called. In this case this module creates and owns the ``scale``/``offset``
        variables.
      - Externally generated ``scale``/``offset``, such as for conditional
        normalization, in which case ``create_*`` should be set to ``False`` and
        then the values fed in at call time.

    NOTE: ``jax.vmap(hk.transform(BatchNorm))`` will update summary statistics and
    normalize values on a per-batch basis; we currently do *not* support
    normalizing across a batch axis introduced by vmap.
    """

    def __init__(
        self,
        create_scale: bool = True,
        create_offset: bool = True,
        decay_rate: float = 0.99,
        eps: float = 1e-5,
        scale_init: Optional[types.Initializer] = None,
        offset_init: Optional[types.Initializer] = None,
        axis: Optional[Sequence[int]] = None,
        cross_replica_axis: Optional[str] = None,
        data_format: str = "channels_last",
        **kwargs
    ):
        """Constructs a BatchNorm module.

        Args:
            create_scale: Whether to include a trainable scaling factor.
            create_offset: Whether to include a trainable offset.
            decay_rate: Decay rate for EMA.
            eps: Small epsilon to avoid division by zero variance. Defaults ``1e-5``,
                as in the paper and Sonnet.
            scale_init: Optional initializer for gain (aka scale). Can only be set
                if ``create_scale=True``. By default, ``1``.
            offset_init: Optional initializer for bias (aka offset). Can only be set
                if ``create_offset=True``. By default, ``0``.
            axis: Which axes to reduce over. The default (``None``) signifies that all
                but the channel axis should be normalized. Otherwise this is a list of
                axis indices which will have normalization statistics calculated.
            cross_replica_axis: If not ``None``, it should be a string representing
                the axis name over which this module is being run within a ``jax.pmap``.
                Supplying this argument means that batch statistics are calculated
                across all replicas on that axis.
            data_format: The data format of the input. Can be either
                ``channels_first``, ``channels_last``, ``N...C`` or ``NC...``. By
                default it is ``channels_last``.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(**kwargs)
        if not create_scale and scale_init is not None:
            raise ValueError("Cannot set `scale_init` if `create_scale=False`")
        if not create_offset and offset_init is not None:
            raise ValueError("Cannot set `offset_init` if `create_offset=False`")

        self.create_scale = create_scale
        self.create_offset = create_offset
        self.eps = eps
        self.scale_init = scale_init or jnp.ones
        self.offset_init = offset_init or jnp.zeros
        self.axis = axis
        self.cross_replica_axis = cross_replica_axis
        self.channel_index = haiku_utils.get_channel_index(data_format)
        self.mean_ema = ExponentialMovingAverage(decay_rate, name="mean_ema")
        self.var_ema = ExponentialMovingAverage(decay_rate, name="var_ema")

    def call(
        self,
        inputs: jnp.ndarray,
        training: tp.Optional[bool] = None,
        test_local_stats: bool = False,
        scale: Optional[jnp.ndarray] = None,
        offset: Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """Computes the normalized version of the input.

        Args:
            inputs: An array, where the data format is ``[..., C]``.
            training: Whether training is currently happening.
            test_local_stats: Whether local stats are used when training=False.
            scale: An array up to n-D. The shape of this tensor must be broadcastable
                to the shape of ``inputs``. This is the scale applied to the normalized
                inputs. This cannot be passed in if the module was constructed with
                ``create_scale=True``.
            offset: An array up to n-D. The shape of this tensor must be broadcastable
                to the shape of ``inputs``. This is the offset applied to the normalized
                inputs. This cannot be passed in if the module was constructed with
                ``create_offset=True``.

        Returns:
            The array, normalized across all but the last dimension.
        """
        inputs = jnp.asarray(inputs, jnp.float32)

        if training is None:
            training = self.is_training()

        if self.create_scale and scale is not None:
            raise ValueError("Cannot pass `scale` at call time if `create_scale=True`.")
        if self.create_offset and offset is not None:
            raise ValueError(
                "Cannot pass `offset` at call time if `create_offset=True`."
            )

        channel_index = self.channel_index
        if channel_index < 0:
            channel_index += inputs.ndim

        if self.axis is not None:
            axis = self.axis
        else:
            axis = [i for i in range(inputs.ndim) if i != channel_index]

        if training or test_local_stats or not hasattr(self.mean_ema, "average"):
            cross_replica_axis = self.cross_replica_axis
            if self.cross_replica_axis:
                mean = jnp.mean(inputs, axis, keepdims=True)
                mean = jax.lax.pmean(mean, cross_replica_axis)
                mean_of_squares = jnp.mean(inputs ** 2, axis, keepdims=True)
                mean_of_squares = jax.lax.pmean(mean_of_squares, cross_replica_axis)
                var = mean_of_squares - mean ** 2
            else:
                mean = jnp.mean(inputs, axis, keepdims=True)
                # This uses E[(X - E[X])^2].
                # TODO(tycai): Consider the faster, but possibly less stable
                # E[X^2] - E[X]^2 method.
                var = jnp.var(inputs, axis, keepdims=True)
        else:
            mean = self.mean_ema.average
            var = self.var_ema.average

        if training or not hasattr(self.mean_ema, "average"):
            self.mean_ema(mean)
            self.var_ema(var)

        w_shape = [1 if i in axis else inputs.shape[i] for i in range(inputs.ndim)]
        w_dtype = jnp.float32

        if self.create_scale:
            scale = self.add_parameter(
                "scale", lambda: self.scale_init(w_shape, w_dtype)
            )
        elif scale is None:
            scale = np.ones([], dtype=w_dtype)

        if self.create_offset:
            offset = self.add_parameter(
                "offset", lambda: self.offset_init(w_shape, w_dtype)
            )
        elif offset is None:
            offset = np.zeros([], dtype=w_dtype)

        inv = scale * jax.lax.rsqrt(var + self.eps)
        output = (inputs - mean) * inv + offset
        return jnp.asarray(output, self.dtype)



# Implementation based on Tensorflow Keras:
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/losses.py#L44-L201

import typing as tp
from abc import abstractmethod
from enum import Enum

import jax.numpy as jnp
import numpy as np
from elegy import types, utils
from numpy.lib.arraysetops import isin


class Reduction(Enum):
    """
    Types of loss reduction.

    Contains the following values:
    * `NONE`: Weighted losses with one dimension reduced (axis=-1, or axis
        specified by loss function). When this reduction type used with built-in
        Keras training loops like `fit`/`evaluate`, the unreduced vector loss is
        passed to the optimizer but the reported loss will be a scalar value.
    * `SUM`: Scalar sum of weighted losses.
    * `SUM_OVER_BATCH_SIZE`: Scalar `SUM` divided by number of elements in losses.
    """

    # AUTO = "auto"
    NONE = "none"
    SUM = "sum"
    SUM_OVER_BATCH_SIZE = "sum_over_batch_size"

    @classmethod
    def all(cls):
        return (
            # cls.AUTO,
            cls.NONE,
            cls.SUM,
            cls.SUM_OVER_BATCH_SIZE,
        )

    @classmethod
    def validate(cls, key):
        if key not in cls.all():
            raise ValueError("Invalid Reduction Key %s." % key)


class Loss:
    """
    Loss base class.

    To be implemented by subclasses:

    * `call()`: Contains the logic for loss calculation.

    Example subclass implementation:

    ```python
    class MeanSquaredError(Loss):
        def call(self, y_true, y_pred):
            return jnp.mean(jnp.square(y_pred - y_true), axis=-1)
    ```

    Please see the [Modules, Losses, and Metrics Guide]
    (https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#losses) for more
    details on this.
    """

    # Methods used for auto docs
    __all__ = ["__init__", "call"]

    def __init__(
        self,
        reduction: tp.Optional[Reduction] = None,
        weight: tp.Optional[float] = None,
        on: tp.Optional[types.IndexLike] = None,
        name: tp.Optional[str] = None,
    ):
        """
        Initializes `Loss` class.

        Arguments:
            reduction: (Optional) Type of `elegy.losses.Reduction` to apply to
                loss. Default value is `SUM_OVER_BATCH_SIZE`. For almost all cases
                this defaults to `SUM_OVER_BATCH_SIZE`.
            weight: Optional weight contribution for the total loss. Defaults to `1`.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
            name: Optional name for the instance, if not provided lower snake_case version
                of the name of the class is used instead.
        """
        self.name = name if name is not None else utils.get_name(self)
        self.weight = weight if weight is not None else 1.0
        self._reduction = (
            reduction if reduction is not None else Reduction.SUM_OVER_BATCH_SIZE
        )
        self._labels_filter = (on,) if isinstance(on, (str, int)) else on
        self._signature_f = self.call

    def __call__(
        self,
        *args,
        **kwargs,
    ):

        if self._labels_filter is not None:
            if "y_true" in kwargs and kwargs["y_true"] is not None:
                for index in self._labels_filter:
                    kwargs["y_true"] = kwargs["y_true"][index]

            if "y_pred" in kwargs and kwargs["y_pred"] is not None:
                for index in self._labels_filter:
                    kwargs["y_pred"] = kwargs["y_pred"][index]

        values = self.call(*args, **kwargs)

        sample_weight: tp.Optional[jnp.ndarray] = kwargs.get("sample_weight", None)

        if isinstance(values, tp.Dict):
            return {
                key: reduce_loss(values, sample_weight, self.weight, self._reduction)
                for key, values in values.items()
            }
        else:
            return reduce_loss(values, sample_weight, self.weight, self._reduction)

    @abstractmethod
    def call(self, *args, **kwargs) -> tp.Any:
        ...


def reduce_loss(values, sample_weight, weight, reduction):

    values: jnp.ndarray = jnp.asarray(values)

    if sample_weight is not None:
        values *= sample_weight

    if reduction == Reduction.NONE:
        loss = values
    elif reduction == Reduction.SUM:
        loss = jnp.sum(values)
    elif reduction == Reduction.SUM_OVER_BATCH_SIZE:
        loss = jnp.sum(values) / jnp.prod(jnp.array(values.shape))
    else:
        raise ValueError(f"Invalid reduction '{reduction}'")

    return loss * weight



import typing as tp

import optax
from elegy import types, utils

from .generalized_optimizer import GeneralizedOptimizer, register_optimizer_for


@register_optimizer_for(optax.GradientTransformation)
class OptaxOptimizer(GeneralizedOptimizer):
    def __init__(self, optimizer: optax.GradientTransformation):
        self.optimizer = optimizer

    def init(
        self, rng: types.RNGSeq, net_params: types.NetParams
    ) -> types.OptimizerStates:
        return self.optimizer.init(net_params)

    def apply(
        self,
        net_params: types.NetParams,
        grads: types.Grads,
        optimizer_states: types.OptimizerStates,
        rng: types.RNGSeq,
    ) -> tp.Tuple[types.NetParams, types.OptimizerStates]:
        updates, optimizer_states = self.optimizer.update(
            grads, optimizer_states, net_params
        )
        net_params = optax.apply_updates(net_params, updates)

        return net_params, optimizer_states



# Implementation based on Tensorflow Keras and Haiku
# Tensorflow: https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/layers/core.py#L127-L224
# Haiku: https://github.com/deepmind/dm-haiku/blob/master/haiku/_src/basic.py#L293#L315

import typing as tp

import haiku as hk
import jax.numpy as jnp
import numpy as np

from elegy import module
from elegy import hooks
from elegy.module import Module


class Dropout(Module):
    """
    Applies Dropout to the input.

    The Dropout layer randomly sets input units to 0 with a frequency of `rate`
    at each step during training time, which helps prevent overfitting.
    Inputs not set to 0 are scaled up by `1/(1 - rate)` such that the sum over
    all inputs is unchanged.

    Note that the Dropout layer only applies when `training` is set to `True`
    such that no values are dropped during inference. When using `model.fit`,
    `training` will be appropriately set to True automatically, and in other
    contexts, you can set the kwarg explicitly to True when calling the layer.

    ### Example
    ```python
    dropout = elegy.nn.Dropout(0.2)
    data = np.arange(10).reshape(5, 2).astype(np.float32)

    print(data)
    # [[0. 1.]
    # [2. 3.]
    # [4. 5.]
    # [6. 7.]
    # [8. 9.]]

    outputs = dropout(data, training=True)

    print(outputs)
    # [[ 0.    1.25]
    # [ 2.5   3.75]
    # [ 5.    6.25]
    # [ 7.5   0.  ]
    # [10.    0.  ]]
    ```
    """

    def __init__(self, rate, **kwargs):
        super().__init__(**kwargs)
        self.rate = rate

    def call(
        self,
        x: jnp.ndarray,
        training: tp.Optional[bool] = None,
        rng: tp.Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """
        Arguments:
            x: The value to be dropped out.
            training: Whether training is currently happening.
            rng: Optional RNGKey.
        Returns:
            x but dropped out and scaled by `1 / (1 - rate)`.
        """
        if training is None:
            training = self.is_training()

        return hk.dropout(
            rng=rng if rng is not None else self.next_key(),
            rate=self.rate if training else 0.0,
            x=x,
        )



import math
from unittest import TestCase

import numpy as np
import torch
from elegy.data.torch_dataloader_adapter import TorchDataLoaderAdapter
from torch.utils.data import DataLoader, TensorDataset


class ArrayDataAdapterTest(TestCase):
    def test_basic(self):
        batch_size = 10
        epochs = 1
        x = np.array(np.random.uniform(size=(100, 32, 32, 3)))
        y = np.array(np.random.uniform(size=(100, 1)))

        dataset = TensorDataset(torch.from_numpy(x), torch.from_numpy(y))
        dataloader = DataLoader(dataset, batch_size=batch_size)

        data_adapter = TorchDataLoaderAdapter(dataloader)

        dataset_length = x.shape[0]
        num_steps = math.ceil(dataset_length / batch_size) * epochs
        iterator_fn = data_adapter.get_dataset()
        for i, batch in zip(range(num_steps), iterator_fn()):
            batch_x, batch_y = batch
            assert batch_x.shape == (batch_size, *x.shape[1:])
            assert batch_y.shape == (batch_size, *y.shape[1:])
            np.testing.assert_array_equal(
                batch_x,
                x[
                    (i * batch_size)
                    % dataset_length : (i * batch_size)
                    % dataset_length
                    + batch_size
                ],
            )

        assert data_adapter.get_size() * batch_size == x.shape[0]
        assert data_adapter.batch_size == batch_size
        assert i == num_steps - 1



import numpy as np
import jax, jax.numpy as jnp
import multiprocessing.pool
import typing as tp
from .data_adapter import DataAdapter
from .utils import is_none_or_empty
import os


__all__ = ["Dataset", "DataLoader"]


_example_usage_docstring = """

Example Usage:
```
class MyDataset(elegy.data.Dataset):
def __len__(self):
    return 128

def __getitem__(self, i):
    #dummy data
    return np.random.random([224, 224, 3]),  np.random.randint(10)

ds     = MyDataset()
loader = elegy.data.DataLoader(ds, batch_size=8, n_workers=8, worker_type='thread', shuffle=True)
model.fit(loader, epochs=10)
```
"""


class Dataset:
    """Abstract base class for datasets. Subclasses should implement the `__getitem__` and `__len__` methods."""  # +_example_usage_docstring

    __all__ = ["__getitem__", "__len__"]

    def __getitem__(self, i: int) -> tp.Any:
        """Abstract method. In a subclass this should return the `i`-th data sample"""
        raise NotImplementedError

    def __len__(self) -> int:
        """Abstract method. In a subclass this should return the number of data samples in the dataset."""
        raise NotImplementedError

    def batch_fn(
        self, list_of_samples: tp.List[tp.Any]
    ) -> tp.Union[jnp.ndarray, tp.Tuple[jnp.ndarray]]:
        """Used by DataLoader to group a list of individual samples into a batch.
        By default tries to stack elements in the samples according to their positiion.
        Can be overridden for more complex use cases.
        """
        return default_batch_fn(list_of_samples)


class DataLoader:
    """Loads samples from a dataset and combines them into batches. Can be directly passed to `Model.fit()`"""  # +_example_usage_docstring

    # TODO: __getitem__  incl slicing e.g. [:5]
    # TODO: n_workers='auto'
    # TODO: timeout parameter

    def __init__(
        self,
        dataset: Dataset,
        batch_size: int,
        n_workers: tp.Optional[int] = 0,
        shuffle: tp.Optional[bool] = False,
        worker_type: tp.Optional[str] = "thread",
        prefetch: tp.Optional[int] = 1,
    ):
        """
        Arguments:
            dataset: The dataset from which to load samples.
                     A subclass of elegy.data.Dataset or an iterable which implements `__getitem__` and `__len__`.
            batch_size: A positive integer specifying how many samples a batch should have.
            n_workers: The number of parallel worker threads or processes which load data from the dataset.
                       A value of 0 (default) means to load data from the main thread.
            shuffle: Whether to load the samples in random order or not. Reshuffles on every epoch if True. Default: False
            worker_type: One of 'thread' (default), 'process', 'spawn', 'fork' or 'forkserver'. Only used if `n_workers`>0.
                         Threads are light-weight but underly the limitations of Python's global interpreter lock.
                         'process' uses the default process type as defined in the `multiprocessing` module.
                         'spawn', 'fork' and 'forkserver' can be used to select a specific process type.
                         For more information consult the Python `multiprocessing` documentation.
            prefetch: Number of batches to prefetch for pipelined execution (Default: 2)
        """
        assert (
            batch_size > 0 and type(batch_size) == int
        ), "batch_size must be a positive integer"
        assert worker_type in ["thread", "process", "spawn", "fork", "forkserver"]
        assert (
            prefetch >= 0 and type(prefetch) == int
        ), "prefetch must be a non-negative integer"

        self.dataset = dataset
        self.batch_size = batch_size
        self.n_workers = n_workers
        self.shuffle = shuffle
        self.worker_type = worker_type
        self.prefetch = prefetch

    def __len__(self) -> int:
        """Returns the number of batches per epoch"""
        return int(np.ceil(len(self.dataset) / self.batch_size))

    def __iter__(self) -> tp.Generator[tp.Any, None, None]:
        """Returns a generator which generates batches of loaded data samples"""
        indices = np.arange(len(self.dataset))
        if self.shuffle:
            np.random.shuffle(indices)

        batched_indices = [
            indices[i:][: self.batch_size]
            for i in range(0, len(indices), self.batch_size)
        ]

        if self.n_workers == 0:
            return mainthread_data_iterator(self.dataset, batched_indices)
        else:
            return MultiProcessIterator(
                self.dataset,
                batched_indices,
                self.n_workers,
                prefetch=self.prefetch,
                worker_type=self.worker_type,
            )


Dataset.__doc__ += _example_usage_docstring
DataLoader.__doc__ += _example_usage_docstring


def default_batch_fn(
    list_of_samples: tp.List[tp.Any],
) -> tp.Union[jnp.ndarray, tp.Tuple[jnp.ndarray]]:
    """Batches individual data samples."""
    assert len(list_of_samples) > 0
    return jax.tree_multimap(lambda *x: jnp.asarray(x), *list_of_samples)


def get_batch_fn(ds: tp.Any) -> tp.Callable:
    """Returns either the batch_fn of the argument if it has one, otherwise `default_batch_fn`
    to allow arrays or datasets that don't inherit from elegy.data.Dataset"""
    return getattr(ds, "batch_fn", default_batch_fn)


def mainthread_data_iterator(
    ds: Dataset, batched_indices: tp.List[tp.List[int]]
) -> tp.Iterable[tp.Any]:
    """Generator that loads datasamples from the data set in the main thread"""
    for batch_of_indices in batched_indices:
        samples = list(map(ds.__getitem__, batch_of_indices))
        yield get_batch_fn(ds)(samples)


class WorkerContext:
    """A namespace to store the Dataset object for each process
    instead of passing it in each iteration which causes re-pickling"""

    _per_process_data = dict()

    @classmethod
    def init(cls, ds, worker_type):
        cls._per_process_data[os.getpid()] = ds
        if worker_type != "thread":
            # disable keyboard interrupts in worker process
            import signal

            signal.signal(signal.SIGINT, signal.SIG_IGN)

    @classmethod
    def get_sample(cls, i):
        ds = cls._per_process_data[os.getpid()]
        return ds[i]


class MultiProcessIterator:
    def __init__(
        self,
        ds: Dataset,
        batched_indices: tp.List[tp.List[int]],
        n_workers: int,
        prefetch: int = 2,
        timeout: int = 10,
        worker_type: str = "thread",
    ):
        self.ds = ds
        self.batched_indices = batched_indices
        self.timeout = timeout

        if worker_type == "thread":
            pool_class = multiprocessing.pool.ThreadPool
        else:
            worker_type = (
                None if worker_type == "process" else worker_type
            )  # None means default
            pool_class = multiprocessing.get_context(worker_type).Pool
        self.worker_pool = pool_class(
            n_workers, initializer=WorkerContext.init, initargs=(ds, worker_type)
        )
        # extra thread to transfer data to the device
        self.data_transfer_worker = multiprocessing.pool.ThreadPool(processes=1)

        self.async_results_queue = []
        for batch_of_indices in batched_indices[:prefetch]:
            self.dispatch_tasks(batch_of_indices)
        self.batched_indices = batched_indices[prefetch:]

    def __iter__(self):
        return self

    def __next__(self):
        if len(self.batched_indices):
            batch_of_indices = self.batched_indices.pop(0)
            self.dispatch_tasks(batch_of_indices)

        if len(self.async_results_queue) == 0:
            raise StopIteration
        async_x = self.async_results_queue.pop(0)

        try:
            x = async_x.get(timeout=self.timeout)
        except KeyboardInterrupt:
            self.shutdown()
            raise
        batch = x
        return batch

    def dispatch_tasks(self, batch_of_indices):
        async_x = self.worker_pool.map_async(WorkerContext.get_sample, batch_of_indices)
        async_x = self.data_transfer_worker.apply_async(
            self.data_transfer_fn, (async_x,)
        )
        self.async_results_queue.append(async_x)

    def shutdown(self):
        self.worker_pool.close()
        for a_result in self.async_results_queue:
            # wait for remaining tasks to finish
            # process workers will hang otherwise
            a_result.wait(timeout=self.timeout)
        self.worker_pool.terminate()
        self.worker_pool.join()

    def __del__(self):
        self.shutdown()

    def data_transfer_fn(self, async_map_result, timeout=10):
        samples = async_map_result.get(timeout)
        batch = get_batch_fn(self.ds)(samples)
        # make sure the batch is transferred to the device
        batch = jax.tree_map(jnp.asarray, batch)
        return batch


class DataLoaderAdapter(DataAdapter):
    @staticmethod
    def can_handle(x, y=None):
        return isinstance(x, DataLoader)

    def __init__(
        self,
        x: DataLoader,
        y=None,
        sample_weights=None,
        **kwargs,
    ):
        # shuffling is performed in the DataLoader
        kwargs.pop("shuffle", None)

        if not is_none_or_empty(y):
            raise ValueError(
                "`y` argument is not supported when using DataLoader as input. The underlying Dataset should return the y values."
            )
        if not is_none_or_empty(sample_weights):
            raise ValueError(
                "`sample_weight` argument is not supported when using DataLoader as input. The underlying Dataset should return the sample weights."
            )

        super().__init__(x, y, **kwargs)
        self._dataloader = x

    def should_recreate_iterator(self):
        return True

    def get_dataset(self):
        dataloader = self._dataloader

        def dataloader_wrapper():
            yield from dataloader

        return dataloader_wrapper

    def get_size(self):
        return len(self._dataloader)

    @property
    def batch_size(self):
        return self.representative_batch_size

    @property
    def representative_batch_size(self):
        return self._dataloader.batch_size

    def has_partial_batch(self):
        return False

    @property
    def partial_batch_size(self):
        return



from elegy import utils
import pytest
from unittest import TestCase


class DIFunctionTests(TestCase):
    def test_positional(self):
        def f(a, b, c):
            return a + b + c

        g = utils.inject_dependencies(f)

        y = g("a", "b", "c")

        assert y == "abc"

    def test_positional_error_missing(self):
        def f(a, b, c):
            return a + b + c

        g = utils.inject_dependencies(f)

        with pytest.raises(TypeError):
            g("a", "b")

    def test_positional_error_remaining(self):
        def f(a, b, c):
            return a + b + c

        g = utils.inject_dependencies(f)

        with pytest.raises(TypeError):
            g("a", "b", "c", "d")

    def test_positional_extras_ok(self):
        def f(a, b, c):
            return a + b + c

        g = utils.inject_dependencies(f)

        y = g("a", "b", "c", d="d")

        assert y == "abc"

    def test_keyword(self):
        def f(a, b, c):
            return a + b + c

        g = utils.inject_dependencies(f)

        y = g(b="b", c="c", a="a")

        assert y == "abc"

    def test_keyword_extras_ok(self):
        def f(a, b, c):
            return a + b + c

        g = utils.inject_dependencies(f)

        y = g(b="b", c="c", a="a", d="d")

        assert y == "abc"

    def test_keyword_error_missing(self):
        def f(a, b, c):
            return a + b + c

        g = utils.inject_dependencies(f)

        with pytest.raises(TypeError):
            g(b="b", c="c")

    def test_mixed(self):
        def f(a, b, c):
            return a + b + c

        g = utils.inject_dependencies(f)

        y = g("a", c="c", b="b")

        assert y == "abc"

    def test_mixed_ignore_duplicated_kwarg_in_arg(self):
        def f(a, b, c):
            return a + b + c

        g = utils.inject_dependencies(f)

        y = g("a", c="c", b="b", a="f")

        assert y == "abc"

    def test_override_defaults(self):
        def f(a, b, c="x"):
            return a + b + c

        g = utils.inject_dependencies(f)

        y = g("a", c="c", b="b")

        assert y == "abc"


class TestMergeStructs:
    def test_basic(self):
        a = dict(x=1)
        b = dict(y=2)

        c = utils.merge_params(a, b)

        assert c == {"x": 1, "y": 2}

    def test_hierarchy(self):
        a = dict(a=dict(x=1))
        b = dict(a=dict(y=2))

        c = utils.merge_params(a, b)

        assert c == {"a": {"x": 1, "y": 2}}

    def test_repeated_leafs(self):
        a = dict(a=dict(x=1))
        b = dict(a=dict(x=2))

        with pytest.raises(ValueError):
            c = utils.merge_params(a, b)

    def test_list(self):
        a = [dict(x=1)]
        b = [dict(y=2)]

        c = utils.merge_params(a, b)

        assert c == [{"x": 1, "y": 2}]

    def test_different_lengths(self):
        a = [dict(x=1)]
        b = []

        with pytest.raises(ValueError):
            c = utils.merge_params(a, b)



# adapted from the flax library https://github.com/google/flax

import pickle
import typing as tp

import jax
import jax.numpy as jnp
import numpy as np
from elegy import hooks, module, nn, types, utils

__all__ = [
    "ResNet",
    "ResNet18",
    "ResNet34",
    "ResNet50",
    "ResNet101",
    "ResNet152",
    "ResNet200",
]


PRETRAINED_URLS = {
    "ResNet18": {
        "url": "https://github.com/poets-ai/elegy-assets/releases/download/resnet18_rev1/ResNet18_ImageNet_rev1.pkl",
        "sha256": "02824ae2f29563add46feff14f40c362ae5f9af3f01ea2edc0812e5ca06ca9ae",
    },
    "ResNet50": {
        "url": "https://github.com/poets-ai/elegy-assets/releases/download/resnet50_rev1/ResNet50_ImageNet_rev1.pkl",
        "sha256": "c69086813ccff6b67b2452daabdf64772f8a7f5c04591e1962185129e18989fc",
    },
}


class ResNetBlock(module.Module):
    """ResNet (identity) block"""

    def __init__(
        self,
        n_filters: int,
        strides: tp.Optional[tp.Tuple[int]] = (1, 1),
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.n_filters = n_filters
        self.strides = strides

    def call(self, x: jnp.ndarray):
        x0 = x
        x = nn.Conv2D(
            self.n_filters,
            (3, 3),
            with_bias=False,
            stride=self.strides,
            dtype=self.dtype,
        )(x)
        x = nn.BatchNormalization(decay_rate=0.9, eps=1e-5)(x)
        x = jax.nn.relu(x)

        x = nn.Conv2D(self.n_filters, (3, 3), with_bias=False, dtype=self.dtype)(x)
        x = nn.BatchNormalization(decay_rate=0.9, eps=1e-5)(x)

        if x0.shape != x.shape:
            x0 = nn.Conv2D(
                self.n_filters,
                (1, 1),
                with_bias=False,
                stride=self.strides,
                dtype=self.dtype,
            )(x0)
            x0 = nn.BatchNormalization(decay_rate=0.9, eps=1e-5)(x0)
        return jax.nn.relu(x0 + x)


class BottleneckResNetBlock(ResNetBlock):
    """ResNet Bottleneck block."""

    def call(self, x: jnp.ndarray):
        x0 = x
        x = nn.Conv2D(self.n_filters, (1, 1), with_bias=False, dtype=self.dtype)(x)
        x = nn.BatchNormalization(decay_rate=0.9, eps=1e-5)(x)
        x = jax.nn.relu(x)
        x = nn.Conv2D(
            self.n_filters,
            (3, 3),
            with_bias=False,
            stride=self.strides,
            dtype=self.dtype,
        )(x)
        x = nn.BatchNormalization(decay_rate=0.9, eps=1e-5)(x)
        x = jax.nn.relu(x)
        x = nn.Conv2D(self.n_filters * 4, (1, 1), with_bias=False, dtype=self.dtype)(x)
        x = nn.BatchNormalization(decay_rate=0.9, eps=1e-5, scale_init=jnp.zeros)(x)

        if x0.shape != x.shape:
            x0 = nn.Conv2D(
                self.n_filters * 4,
                (1, 1),
                with_bias=False,
                stride=self.strides,
                dtype=self.dtype,
            )(x0)
            x0 = nn.BatchNormalization(decay_rate=0.9, eps=1e-5)(x0)
        return jax.nn.relu(x0 + x)


class ResNet(module.Module):
    """A generic ResNet V1 architecture that can be customized for non-standard configurations
    Original Paper: [Deep Residual Learning for Image Recognition](https://arxiv.org/abs/1512.03385)
    """

    __all__ = ["__init__", "call"]

    def __init__(
        self,
        stages: tp.List[int],
        block_type: tp.Union[tp.Type[ResNetBlock], tp.Type[BottleneckResNetBlock]],
        lowres: bool = False,
        weights: tp.Optional[str] = None,
        dtype: tp.Optional[tp.Any] = jnp.float32,
        *args,
        **kwargs,
    ):
        """
        Arguments:
            stages: A list of integers representing the number of blocks in each stage.
                    e.g: [3, 4, 6, 3] for a ResNet50
            block_type: Which ResNet block type to use.
            lowres: Optional, whether to use the low resolution version
                    as described in subsection 4.2 of the orignal paper.
                    This version is better suited for datasets like CIFAR10. (Default: False)
            weights: One of None (random initialization) or a path to a weights file
            dtype: Optional dtype of the convolutions and linear operations,
                    either jnp.float32 (default) or jnp.float16 for mixed precision.
        """

        super().__init__(*args, **kwargs)
        self.stages = stages
        self.block_type = block_type
        self.lowres = lowres

        if weights is not None:
            if weights.endswith(".pkl"):
                parameters = pickle.load(open(weights, "rb"))
            elif weights == "imagenet":
                clsname = self.__class__.__name__
                urldict = PRETRAINED_URLS.get(clsname, None)
                if urldict is None:
                    raise ValueError(f"No pretrained weights for {clsname} available")
                fname = utils.download_file(urldict["url"], sha256=urldict["sha256"])
                parameters = pickle.load(open(fname, "rb"))
            else:
                raise ValueError("Unknown weights value: ", weights)

            x = np.empty([0, 224, 224, 3], dtype=self.dtype)
            # quick but dirty module initialization
            jax.eval_shape(self.init(rng=types.RNGSeq(42)), x)

            self.set_default_parameters(parameters)

    def call(self, x: jnp.ndarray):
        x = nn.Conv2D(
            64,
            (7, 7) if not self.lowres else (3, 3),
            stride=(2, 2) if not self.lowres else (1, 1),
            padding="SAME",
            with_bias=False,
            dtype=self.dtype,
        )(x)
        x = nn.BatchNormalization(decay_rate=0.9, eps=1e-5)(x)
        x = module.to_module(jax.nn.relu)()(x)

        if not self.lowres:
            x = nn.MaxPool(
                window_shape=(1, 3, 3, 1), strides=(1, 2, 2, 1), padding="SAME"
            )(x)
        for i, block_size in enumerate(self.stages):
            for j in range(block_size):
                strides = (2, 2) if i > 0 and j == 0 else (1, 1)
                x = self.block_type(64 * 2 ** i, strides=strides, dtype=self.dtype)(x)
        GAP = lambda x: jnp.mean(x, axis=(1, 2))
        x = module.to_module(GAP)(name="global_average_pooling")(x)
        x = nn.Linear(1000, dtype=self.dtype)(x)
        to_float32 = lambda x: jnp.asarray(x, jnp.float32)
        x = module.to_module(to_float32)(name="to_float32")(x)
        return x


class ResNet18(ResNet):
    def __init__(
        self,
        lowres: bool = False,
        weights: tp.Optional[str] = None,
        dtype: tp.Optional[tp.Any] = jnp.float32,
        *args,
        **kwargs,
    ):
        super().__init__(
            stages=[2, 2, 2, 2],
            block_type=ResNetBlock,
            lowres=lowres,
            weights=weights,
            dtype=dtype,
            *args,
            **kwargs,
        )


class ResNet34(ResNet):
    def __init__(
        self,
        lowres: bool = False,
        weights: tp.Optional[str] = None,
        dtype: tp.Optional[tp.Any] = jnp.float32,
        *args,
        **kwargs,
    ):
        super().__init__(
            stages=[3, 4, 6, 3],
            block_type=ResNetBlock,
            lowres=lowres,
            weights=weights,
            dtype=dtype,
            *args,
            **kwargs,
        )


class ResNet50(ResNet):
    def __init__(
        self,
        lowres: bool = False,
        weights: tp.Optional[str] = None,
        dtype: tp.Optional[tp.Any] = jnp.float32,
        *args,
        **kwargs,
    ):
        super().__init__(
            stages=[3, 4, 6, 3],
            block_type=BottleneckResNetBlock,
            lowres=lowres,
            weights=weights,
            dtype=dtype,
            *args,
            **kwargs,
        )


class ResNet101(ResNet):
    def __init__(
        self,
        lowres: bool = False,
        weights: tp.Optional[str] = None,
        dtype: tp.Optional[tp.Any] = jnp.float32,
        *args,
        **kwargs,
    ):
        super().__init__(
            stages=[3, 4, 23, 3],
            block_type=BottleneckResNetBlock,
            lowres=lowres,
            weights=weights,
            dtype=dtype,
            *args,
            **kwargs,
        )


class ResNet152(ResNet):
    def __init__(
        self,
        lowres: bool = False,
        weights: tp.Optional[str] = None,
        dtype: tp.Optional[tp.Any] = jnp.float32,
        *args,
        **kwargs,
    ):
        super().__init__(
            stages=[3, 8, 36, 3],
            block_type=BottleneckResNetBlock,
            lowres=lowres,
            weights=weights,
            dtype=dtype,
            *args,
            **kwargs,
        )


class ResNet200(ResNet):
    def __init__(
        self,
        lowres: bool = False,
        weights: tp.Optional[str] = None,
        dtype: tp.Optional[tp.Any] = jnp.float32,
        *args,
        **kwargs,
    ):
        super().__init__(
            stages=[3, 24, 36, 3],
            block_type=BottleneckResNetBlock,
            lowres=lowres,
            weights=weights,
            dtype=dtype,
            *args,
            **kwargs,
        )


_resnet__init___docstring = """
Instantiates the {} architecture from [Deep Residual Learning for Image Recognition](https://arxiv.org/abs/1512.03385)

Arguments:
    lowres: Optional, whether to use the low resolution version
            as described in subsection 4.2 of the orignal paper.
            This version is better suited for datasets like CIFAR10. (Default: False)
    weights: One of None (random initialization), 'imagenet' (automatic download of
              weights pretrained on ImageNet) or a path to a weights file
    dtype: Optional dtype of the convolutions and linear operations, 
           either jnp.float32 (default) or jnp.float16 for mixed precision.
"""

ResNet18.__init__.__doc__ = _resnet__init___docstring.format("ResNet18")
ResNet34.__init__.__doc__ = _resnet__init___docstring.format("ResNet34")
ResNet50.__init__.__doc__ = _resnet__init___docstring.format("ResNet50")
ResNet101.__init__.__doc__ = _resnet__init___docstring.format("ResNet101")
ResNet152.__init__.__doc__ = _resnet__init___docstring.format("ResNet152")
ResNet200.__init__.__doc__ = _resnet__init___docstring.format("ResNet200")



import typing as tp

import jax.numpy as jnp
import numpy as np
import optax

from elegy import module, types, utils
from elegy.generalized_optimizer.generalized_optimizer import GeneralizedOptimizer


class LRScheduler(types.Protocol):
    def __call__(
        self, step: jnp.ndarray, epoch: tp.Optional[jnp.ndarray] = None
    ) -> jnp.ndarray:
        ...


class Optimizer(GeneralizedOptimizer):
    r"""A Module that wraps around `optax` optimizers."""

    def __init__(
        self,
        *optimizer: optax.GradientTransformation,
        lr_schedule: tp.Optional[LRScheduler] = None,
        steps_per_epoch: tp.Union[int, jnp.ndarray, np.ndarray, None] = None,
        **kwargs,
    ):
        r"""
        Arguments:
            optimizer: An optax `GradientTransformation` object, if more than one is passed via `*args` then they are
                grouped using `optax.chain`.
            lr_schedule: A optional callable of the form `def lr_schedule(step: int, epoch: Optional[int]) -> float` that
                returns the learning rate schedule at each time step. If `steps_per_epoch` is given then epoch is calculated,
                else epoch is None.
            steps_per_epoch: The number of steps to in an epoch, needed to caculate `epoch` from `step`.
        """

        if len(optimizer) == 0:
            raise ValueError("Must pass atleast 1 optimizer, got 0")

        elif lr_schedule is not None:
            # do this to preserve reference after re-assign latter
            base_schedule = lr_schedule

            def lr_schedule_(step: jnp.ndarray) -> jnp.ndarray:
                epoch: tp.Any = (
                    step // steps_per_epoch if steps_per_epoch is not None else None
                )

                return base_schedule(step, epoch)

            optimizer = optax.chain(
                *optimizer,
                optax.scale_by_schedule(lr_schedule_),
            )

            lr_schedule = lr_schedule_

        elif len(optimizer) == 1:
            optimizer = optimizer[0]
        else:
            optimizer = optax.chain(*optimizer)

        self.optimizer = optimizer
        self.lr_schedule = lr_schedule

    def init(
        self, rng: types.RNGSeq, net_params: types.NetParams
    ) -> types.OptimizerStates:
        return self.optimizer.init(net_params)

    def apply(
        self,
        net_params: types.NetParams,
        grads: types.Grads,
        optimizer_states: types.OptimizerStates,
        rng: types.RNGSeq,
    ) -> tp.Tuple[types.NetParams, types.OptimizerStates]:
        updates, optimizer_states = self.optimizer.update(
            grads, optimizer_states, net_params
        )
        net_params = optax.apply_updates(net_params, updates)

        return net_params, optimizer_states

    def current_lr(
        self, optimizer_states: types.OptimizerStates
    ) -> tp.Optional[jnp.ndarray]:
        """Returns the learning rate scaled by schedule(s) that will be used for the next training step"""

        if (
            not isinstance(optimizer_states, types.Uninitialized)
            and self.lr_schedule is not None
        ):
            step = optimizer_states[-1].count
            return self.lr_schedule(step)



# Implementation based on tf.keras.engine.training.py
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/engine/training.py

import pickle
import typing as tp
from copy import copy
from pathlib import Path

import jax.numpy as jnp
import numpy as np
from elegy import types
from elegy.callbacks import Callback, CallbackList, History
from elegy.data import (
    DataHandler,
    map_append,
    map_structure,
    train_validation_split,
    unpack_x_y_sample_weight,
)
from elegy.model.model_core import ModelCore, PredStep, TestStep
from tabulate import tabulate

# from elegy.module import Module


__all__ = ["Model", "load"]


class ModelBase(ModelCore):
    """
    `Model` is tasked with performing training, evaluation, and inference for a given
    `elegy.Module` or `haiku.Module`.

    To create a `Model` you first have to define its architecture in a `Module`:

    ```python
    >>> import elegy, jax
    >>> import jax.numpy as jnp

    >>> class MLP(elegy.Module):
    ...     def call(self, x: jnp.ndarray) -> jnp.ndarray:
    ...         x = elegy.nn.Flatten()(x)
    ...         x = elegy.nn.Linear(5)(x)
    ...         x = jax.nn.relu(x)
    ...         x = elegy.nn.Linear(2)(x)
    ...         return x

    >>> mlp = MLP()
    >>> x = jnp.ones(shape=[10, 2])

    >>> y_pred, collections = mlp.init(rng=elegy.RNGSeq(42))(x)
    >>> y_pred.shape
    (10, 2)

    ```


    You can pass use `Module` with the Model API:
    ```python
    model = elegy.Model(
        module=MLP(),
        loss=[
            elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
            elegy.regularizers.GlobalL2(l=1e-5),
        ],
        metrics=elegy.metrics.SparseCategoricalAccuracy(),
        optimizer=optax.rmsprop(1e-3),
    )
    ```

    Once the model is created, you can train the model with `model.fit()`, or use the model
    to do prediction with `model.predict()`.
    Checkout [Getting Started](https://poets-ai.github.io/elegy/getting-started) for
    additional details.

    Model supports defining + monitoring custom learning rate schedules by passing an instance of `elegy.Optimizer` instead of
    an `optax` object:

    ```python
    model = elegy.Model(
        module=MLP(n1=3, n2=1),
        loss=elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
        metrics=elegy.metrics.SparseCategoricalAccuracy(),
        optimizer=elegy.Optimizer(
            optax.adam(1.0), # <---- important to set this to 1.0
            lr_schedule=lambda step, epoch: 1 / (epoch * 100 + step),
            steps_per_epoch=1000,
        ),
        run_eagerly=True,
    )

    history = model.fit(
        ...
    )

    assert "lr" in history.history
    ```
    Notice how we set the learning rate parameter of the `adam` optimizer to `1.0`, this is necessary if you want the logged `lr`
    be closer to the "actual" learning rate because we implement this feature by chaining an additional `optax.scale_by_schedule`
    at the end.
    """

    __all__ = [
        "evaluate",
        "fit",
        "load",
        "predict",
        "predict_on_batch",
        "reset",
        "reset_metrics",
        "save",
        "summary",
        "test_on_batch",
        "train_on_batch",
        "full_state",
        "parameters",
        "states",
    ]

    def pred_step(self, *args, **kwargs):
        raise types.MissingMethod()

    def test_step(self, *args, **kwargs):
        raise types.MissingMethod()

    def train_step(self, *args, **kwargs):
        raise types.MissingMethod()

    def fit(
        self,
        x: tp.Union[
            np.ndarray,
            tp.Mapping[str, np.ndarray],
            tp.Tuple[np.ndarray],
            tp.Iterable,
            None,
        ] = None,
        y: tp.Union[
            np.ndarray,
            tp.Mapping[str, np.ndarray],
            tp.Tuple[np.ndarray],
            None,
        ] = None,
        batch_size: tp.Optional[int] = None,
        epochs: int = 1,
        verbose: int = 1,
        callbacks: tp.Union[tp.List[Callback], CallbackList, None] = None,
        validation_split: float = 0.0,
        validation_data: tp.Union[tp.Tuple, tp.Iterable, None] = None,
        shuffle: bool = True,
        class_weight: tp.Optional[tp.Mapping[str, float]] = None,
        sample_weight: tp.Optional[np.ndarray] = None,
        initial_epoch: int = 0,
        steps_per_epoch: tp.Optional[int] = None,
        validation_steps: tp.Optional[int] = None,
        validation_batch_size: tp.Optional[int] = None,
        validation_freq: int = 1,
    ) -> History:
        """
        Trains the model for a fixed number of epochs (iterations on a dataset).

        Arguments:
            x: Input data. It could be:

                - A Numpy or Jax array (or array-like), or a list of arrays
                    (in case the model has multiple inputs).
                - A dict mapping input names to the corresponding arrays,
                    if the model has named inputs.
                - A generator returning `(inputs,)`, `(inputs, targets)`
                    or `(inputs, targets, sample_weights)`.

                A more detailed description of unpacking behavior for generator type
                is given below.
            y: Target data. Like the input data `x`,
                it could be either Numpy or Jax array(s).
                It should be consistent with `x`. If `x` is a generator,
                `y` should not be specified (since targets will be obtained from `x`).
            batch_size: Integer or `None`.
                Number of samples per gradient update.
                If unspecified, `batch_size` will default to 32.
                Do not specify the `batch_size` if your data is in the
                form of generator (since they generate batches).
            epochs: Integer. Number of epochs to train the model.
                An epoch is an iteration over the entire `x` and `y`
                data provided.
                Note that in conjunction with `initial_epoch`,
                `epochs` is to be understood as "final epoch".
                The model is not trained for a number of iterations
                given by `epochs`, but merely until the epoch
                of index `epochs` is reached.
            verbose: 0, 1, 2 or 3. Verbosity mode.
                0 = silent, 1 = progress bar, 2 = one line per epoch 3 = table.
                Note that the progress bar is not particularly useful when
                logged to a file, so verbose=2 is recommended when not running
                interactively (eg, in a production environment).
            callbacks: List of [elegy.callbacks.callback.Callback][] instances.
                List of callbacks to apply during training.
            validation_split: Float between 0 and 1.
                Fraction of the training data to be used as validation data.
                The model will set apart this fraction of the training data,
                will not train on it, and will evaluate
                the loss and any model metrics
                on this data at the end of each epoch.
                The validation data is selected from the last samples
                in the `x` and `y` data provided, before shuffling. This argument is
                not supported when `x` is a generator.
            validation_data: Data on which to evaluate
                the loss and any model metrics at the end of each epoch.
                The model will not be trained on this data.
                `validation_data` will override `validation_split`.
                `validation_data` could be:

                - tuple `(x_val, y_val)` of Numpy/Jax arrays, list of arrays or mappings
                - tuple `(x_val, y_val, val_sample_weights)` of Numpy/Jax arrays, list
                of arrays or mappings
                - generator

                For the first two cases, `batch_size` must be provided.
                For the last case, `validation_steps` should be provided, and should
                follow the same convention for yielding data as `x`.
                Note that `validation_data` does not support all the data types that
                are supported in `x`, eg, dict.
            shuffle: Boolean (whether to shuffle the training data
                before each epoch). This argument is ignored
                when `x` is a generator. Has no effect when `steps_per_epoch` is not `None`.
            class_weight: Optional dictionary mapping class indices (integers)
                to a weight (float) value, used for weighting the loss function
                (during training only).
                This can be useful to tell the model to
                "pay more attention" to samples from
                an under-represented class.
            sample_weight: Optional Numpy/Jax array of weights for
                the training samples, used for weighting the loss function
                (during training only). You can either pass a flat (1D)
                Numpy array with the same length as the input samples
                (1:1 mapping between weights and samples). This argument is not
                supported when `x` is generator, instead provide the sample_weights
                as the third element of `x`.
            initial_epoch: Integer.
                Epoch at which to start training
                (useful for resuming a previous training run).
            steps_per_epoch: Integer or `None`.
                Total number of steps (batches of samples)
                before declaring one epoch finished and starting the
                next epoch. When training with input arrays such as
                jax data arrays, the default `None` is equal to
                the number of samples in your dataset divided by
                the batch size, or 1 if that cannot be determined.
                When passing a generator, you must specify the
                `steps_per_epoch` argument. This argument is not supported with
                array inputs.
            validation_steps: Only relevant if `validation_data` is provided and
                is a generator. Total number of steps (batches of
                samples) to draw before stopping when performing validation
                at the end of every epoch. If 'validation_steps' is None, validation
                will run until the `validation_data` dataset is exhausted. In the
                case of an infinitely repeated dataset, it will run into an
                infinite loop. If 'validation_steps' is specified and only part of
                the dataset will be consumed, the evaluation will start from the
                beginning of the dataset at each epoch. This ensures that the same
                validation samples are used every time.
            validation_batch_size: Integer or `None`.
                Number of samples per validation batch.
                If unspecified, will default to `batch_size`.
                Do not specify the `validation_batch_size` if your data is in the
                form of generators (since they generate batches).
            validation_freq: Only relevant if validation data is provided. Integer
                or `collections_abc.Container` instance (e.g. list, tuple, etc.).
                If an integer, specifies how many training epochs to run before a
                new validation run is performed, e.g. `validation_freq=2` runs
                validation every 2 epochs. If a Container, specifies the epochs on
                which to run validation, e.g. `validation_freq=[1, 2, 10]` runs
                validation at the end of the 1st, 2nd, and 10th epochs.

        Unpacking behavior for iterator-like inputs:

        A common pattern is to pass a generator, which will in fact
        yield not only features (x) but optionally targets (y) and sample weights.
        Elegy requires that the output of such iterator-likes be unambiguous. The
        iterator should return a tuple of length 1, 2, or 3, where the optional
        second and third elements will be used for y and sample_weight
        respectively. Any other type provided will be wrapped in a length one
        tuple, effectively treating everything as 'x'. When yielding dicts, they
        should still adhere to the top-level tuple structure.
        e.g. `({"x0": x0, "x1": x1}, y)`. Elegy will not attempt to separate
        features, targets, and weights from the keys of a single dict.

        A notable unsupported data type is the namedtuple. The reason is that
        it behaves like both an ordered datatype (tuple) and a mapping
        datatype (dict). So given a namedtuple of the form:
            `namedtuple("example_tuple", ["y", "x"])`
        it is ambiguous whether to reverse the order of the elements when
        interpreting the value. Even worse is a tuple of the form:
            `namedtuple("other_tuple", ["x", "y", "z"])`
        where it is unclear if the tuple was intended to be unpacked into x, y,
        and sample_weight or passed through as a single element to `x`. As a
        result the data processing code will simply raise a ValueError if it
        encounters a namedtuple. (Along with instructions to remedy the issue.)

        Returns:
            A `History` object. Its `History.history` attribute is
            a record of training loss values and metrics values
            at successive epochs, as well as validation loss values
            and validation metrics values (if applicable).
        Raises:
            ValueError: In case of mismatch between the provided input data
                and what the model expects.
        """
        if x is None:
            x = {}

        if validation_split:
            # Create the validation data using the training data. Only supported for
            # `Jax Numpy` and `NumPy` input.
            (x, y, sample_weight), validation_data = train_validation_split(
                (x, y, sample_weight), validation_split=validation_split, shuffle=False
            )

        self.stop_training = False
        data_handler = DataHandler(
            x=x,
            y=y,
            sample_weight=sample_weight,
            batch_size=batch_size,
            steps_per_epoch=steps_per_epoch,
            initial_epoch=initial_epoch,
            epochs=epochs,
            shuffle=shuffle,
            class_weight=class_weight,
        )
        # Container that configures and calls `tf.keras.Callback`s.
        if not isinstance(callbacks, CallbackList):
            callbacks = CallbackList(
                callbacks,
                add_history=True,
                add_progbar=verbose != 0,
                model=self,
                verbose=verbose,
                epochs=epochs,
                steps=data_handler.inferred_steps,
            )

        callbacks.on_train_begin()
        # data_handler._initial_epoch = (  # pylint: disable=protected-access
        #     self._maybe_load_initial_epoch_from_ckpt(initial_epoch))

        for epoch, iterator in data_handler.enumerate_epochs():
            self.reset_metrics()
            callbacks.on_epoch_begin(epoch)
            logs = {}
            with data_handler.catch_stop_iteration():
                for step in data_handler.steps():
                    callbacks.on_train_batch_begin(step)
                    batch = next(iterator)
                    # sample_weight = batch[2] if len(batch) == 3 else None
                    x_batch, y_batch, sample_weight = unpack_x_y_sample_weight(batch)

                    tmp_logs = self.train_on_batch(
                        x=x_batch,
                        y=y_batch,
                        sample_weight=sample_weight,
                        class_weight=class_weight,
                    )
                    tmp_logs.update({"size": data_handler.batch_size})
                    # print(epoch, step, tmp_logs["accuracy"], batch[0].shape)

                    logs = tmp_logs
                    callbacks.on_train_batch_end(step, logs)

            epoch_logs = copy(logs)
            epoch_logs.update({"size": data_handler.batch_size})

            # Run validation.
            if validation_data and self._should_eval(epoch, validation_freq):
                val_x, val_y, val_sample_weight = unpack_x_y_sample_weight(
                    validation_data
                )
                try:
                    val_logs = self.evaluate(
                        x=val_x,
                        y=val_y,
                        sample_weight=val_sample_weight,
                        batch_size=validation_batch_size or batch_size,
                        steps=validation_steps,
                        callbacks=callbacks,
                        # return_dict=True,
                    )

                    val_logs = {"val_" + name: val for name, val in val_logs.items()}
                    epoch_logs.update(val_logs)
                except (types.MissingMethod, types.MissingModule) as e:
                    pass

            callbacks.on_epoch_end(epoch, epoch_logs)
            # print(
            #     f"epoch: {epoch} - "
            #     + " - ".join(f"{key}: {value:.3f}" for key, value in epoch_logs.items())
            # )
            if self.stop_training:
                break

        callbacks.on_train_end()

        return self.history

    def evaluate(
        self,
        x: tp.Union[
            np.ndarray,
            tp.Mapping[str, np.ndarray],
            tp.Tuple[np.ndarray],
            tp.Iterable,
        ],
        y: tp.Union[
            jnp.ndarray,
            np.ndarray,
            tp.Mapping[str, np.ndarray],
            tp.Tuple[np.ndarray],
            None,
        ] = None,
        verbose: int = 1,
        batch_size: tp.Optional[int] = None,
        sample_weight: tp.Optional[np.ndarray] = None,
        steps: tp.Optional[int] = None,
        callbacks: tp.Union[tp.List[Callback], CallbackList, None] = None,
    ) -> types.Logs:
        """Returns the loss value & metrics values for the model in test mode.
        Computation is done in batches.

        Arguments:
            x: Input data. It could be:

                - A Numpy or Jax array (or array-like), or a list of arrays
                    (in case the model has multiple inputs).
                - A dict mapping input names to the corresponding arrays,
                    if the model has named inputs.
                - A generator returning `(inputs,)`, `(inputs, targets)`
                    or `(inputs, targets, sample_weights)`.

                A more detailed description of
                unpacking behavior for iterator types generator
                is given in the `Unpacking behavior for iterator-like inputs` section
                of `Model.fit`.
            y: Target data. Like the input data `x`,
                it could be either Numpy or Jax array(s).
                It should be consistent with `x`. If `x` is a generator,
                `y` should not be specified (since targets will be obtained from `x`).
            verbose: 0, 1, or 2. Verbosity mode.
                0 = silent, 1 = progress bar, 2 = one line per epoch.
            batch_size: Integer or `None`.
                Number of samples per gradient update.
                If unspecified, `batch_size` will default to 32.
                Do not specify the `batch_size` if your data is in the
                form of generator (since they generate batches).
            sample_weight: Optional Numpy/Jax array of weights for
                the training samples, used for weighting the loss function
                (during training only). You can either pass a flat (1D)
                Numpy array with the same length as the input samples
                (1:1 mapping between weights and samples). This argument is not
                supported when `x` is generator, instead provide the sample_weights
                as the third element of `x`.
            steps: Integer or `None`. Total number of steps (batches of samples)
                before declaring the evaluation round finished. Ignored with the
                default value of `None`. This
                argument is not supported with array inputs.
            callbacks: List of [elegy.callbacks.callback.Callback][] instances.
                List of callbacks to apply during training.

        See the discussion of `Unpacking behavior for iterator-like inputs` for
        [`Model.fit`][elegy.model.model.Model.fit].

        Returns:
            A dictionary for mapping the losses and metrics names to the values obtained.
        Raises:
            ValueError: in case of invalid arguments.
        """

        data_handler = DataHandler(
            x=x,
            y=y,
            sample_weight=sample_weight,
            batch_size=batch_size,
            steps_per_epoch=steps,
            initial_epoch=0,
            epochs=1,
            shuffle=False,
            training=False,
        )

        # Container that configures and calls `tf.keras.Callback`s.
        if not isinstance(callbacks, CallbackList):
            callbacks = CallbackList(
                callbacks,
                add_history=True,
                add_progbar=verbose != 0,
                model=self,
                verbose=verbose,
                epochs=1,
                steps=data_handler.inferred_steps,
            )

        callbacks.on_test_begin()

        logs = {}
        for _, iterator in data_handler.enumerate_epochs():
            self.reset_metrics()
            with data_handler.catch_stop_iteration():
                for step in data_handler.steps():
                    callbacks.on_test_batch_begin(step)
                    batch = next(iterator)
                    x_batch, y_batch, sample_weight = unpack_x_y_sample_weight(batch)

                    tmp_logs = self.test_on_batch(
                        x=x_batch,
                        y=y_batch,
                        sample_weight=sample_weight,
                    )
                    tmp_logs.update({"size": data_handler.batch_size})
                    logs = tmp_logs
                    callbacks.on_test_batch_end(step, logs)

        callbacks.on_test_end()

        return logs

    def predict(
        self,
        x: tp.Union[
            np.ndarray,
            tp.Mapping[str, np.ndarray],
            tp.Tuple[np.ndarray],
            tp.Iterable,
        ],
        verbose: int = 0,
        batch_size: tp.Optional[int] = None,
        steps: tp.Optional[int] = None,
        callbacks: tp.Union[tp.List[Callback], CallbackList, None] = None,
    ) -> tp.Any:
        """Generates output predictions for the input samples.
        Computation is done in batches.

        Arguments:
            x: Input data. It could be:

                - A Numpy or Jax array (or array-like), or a list of arrays
                    (in case the model has multiple inputs).
                - A dict mapping input names to the corresponding arrays,
                    if the model has named inputs.
                - A generator returning `(inputs,)`, `(inputs, targets)`
                    or `(inputs, targets, sample_weights)`.

                A more detailed description of
                unpacking behavior for iterator types generator
                is given in the `Unpacking behavior for iterator-like inputs` section
                of `Model.fit`.
            batch_size: Integer or `None`.
                Number of samples per batch.
                If unspecified, `batch_size` will default to 32.
                Do not specify the `batch_size` if your data is in the
                form of generators (since they generate batches).
            verbose: Verbosity mode, 0 or 1.
            steps: Total number of steps (batches of samples)
                before declaring the prediction round finished.
                Ignored with the default value of `None`.
            callbacks: List of [elegy.callbacks.callback.Callback][] instances.
                List of callbacks to apply during training.

        See the discussion of `Unpacking behavior for iterator-like inputs` for
        [`Model.fit`][elegy.model.model.Model.fit].
        Note that Model.predict uses the same interpretation rules as
        `Model.fit` and `Model.evaluate`, so inputs must be unambiguous for all
        three methods.
        Returns:
            Numpy array(s) of predictions.
        Raises:
            ValueError: In case of mismatch between the provided
                input data and the model's expectations,
                or in case a stateful model receives a number of samples
                that is not a multiple of the batch size.
        """

        outputs = None

        data_handler = DataHandler(
            x=x,
            batch_size=batch_size,
            steps_per_epoch=steps,
            initial_epoch=0,
            epochs=1,
            shuffle=False,
        )

        # Container that configures and calls `tf.keras.Callback`s.
        if not isinstance(callbacks, CallbackList):
            callbacks = CallbackList(
                callbacks,
                add_history=True,
                add_progbar=verbose != 0,
                model=self,
                verbose=verbose,
                epochs=1,
                steps=data_handler.inferred_steps,
            )

        callbacks.on_predict_begin()

        for _, iterator in data_handler.enumerate_epochs():
            self.reset_metrics()
            with data_handler.catch_stop_iteration():
                for step in data_handler.steps():
                    callbacks.on_predict_batch_begin(step)
                    batch = next(iterator)
                    tmp_batch_outputs = self.predict_on_batch(x=batch[0])
                    batch_outputs = tmp_batch_outputs

                    if outputs is None:
                        outputs = map_structure(
                            lambda batch_output: [batch_output], batch_outputs
                        )
                    else:

                        outputs = map_structure(
                            map_append,
                            outputs,
                            batch_outputs,
                        )

                    callbacks.on_predict_batch_end(
                        step,
                        {"outputs": batch_outputs, "size": data_handler.batch_size},
                    )

        callbacks.on_predict_end()

        all_outputs = map_structure(jnp.concatenate, outputs)

        return all_outputs

    def _should_eval(self, epoch, validation_freq):
        epoch = epoch + 1  # one-index the user-facing epoch.
        if isinstance(validation_freq, int):
            return epoch % validation_freq == 0
        elif isinstance(validation_freq, list):
            return epoch in validation_freq
        else:
            raise ValueError("Expected `validation_freq` to be a list or int.")


def load(path: tp.Union[str, Path]) -> ModelBase:
    """
    Loads a model from disk.

    This function will restore both the model architecture,
    that is, its `Model` class instance, along with all of its
    parameters, states, and optimizer states.

    Example:

    ```python
    import elegy

    model.save('my_model')  # creates folder at 'my_model'
    del model  # deletes the existing model

    # returns a model identical to the previous one
    model = elegy.model.load('my_model')
    ```

    Arguments:
        path: path to a saved model's directory.
        include_optimizer: If True, loads optimizer's state if available.

    Raises:
        OSError: in case the model was not found or could not be
            loaded from disk successfully.
    """
    if isinstance(path, str):
        path = Path(path)

    try:
        model_bytes = (path / "model.pkl").read_bytes()
    except BaseException as e:
        raise OSError(f"Could not load the model. Got exception: {e}")

    model: ModelBase = pickle.loads(model_bytes)
    model.load(path)

    return model



from unittest import TestCase

import jax.numpy as jnp
import tensorflow.keras as tfk
import numpy as np

import elegy


class BinaryCrossentropyTest(TestCase):
    #
    def test_example(self):
        y_true = np.array([[1], [1], [0], [0]])
        y_pred = np.array([[1], [1], [0], [0]])
        m = elegy.metrics.binary_accuracy(y_true, y_pred)
        assert m.shape == (4,)

        m = elegy.metrics.BinaryAccuracy()
        result = m.call_with_defaults()(
            y_true=np.array([[1], [1], [0], [0]]),
            y_pred=np.array([[0.98], [1], [0], [0.6]]),
        )
        assert result == 0.75

        m = elegy.metrics.BinaryAccuracy()
        result = m.call_with_defaults()(
            y_true=np.array([[1], [1], [0], [0]]),
            y_pred=np.array([[0.98], [1], [0], [0.6]]),
            sample_weight=np.array([1, 0, 0, 1]),
        )
        assert result == 0.5

    #
    def test_compatibility(self):

        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.float32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))
        sample_weight = np.random.uniform(0, 1, size=(5, 6))

        assert np.allclose(
            tfk.metrics.BinaryAccuracy()(y_true, y_pred),
            elegy.metrics.BinaryAccuracy().call_with_defaults()(y_true, y_pred),
        )

        assert np.allclose(
            tfk.metrics.BinaryAccuracy(threshold=0.3)(y_true, y_pred),
            elegy.metrics.BinaryAccuracy(threshold=0.3).call_with_defaults()(
                y_true, y_pred
            ),
        )

        assert np.allclose(
            tfk.metrics.BinaryAccuracy(threshold=0.3)(
                y_true, y_pred, sample_weight=sample_weight
            ),
            elegy.metrics.BinaryAccuracy(threshold=0.3).call_with_defaults()(
                y_true, y_pred, sample_weight=sample_weight
            ),
        )

    def test_cumulative(self):

        tm = tfk.metrics.BinaryAccuracy(threshold=0.3)
        em = elegy.metrics.BinaryAccuracy(threshold=0.3)

        # 1st run
        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.float32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))
        sample_weight = np.random.uniform(0, 1, size=(5, 6))

        assert np.allclose(
            tm(y_true, y_pred, sample_weight=sample_weight),
            em.call_with_defaults()(y_true, y_pred, sample_weight=sample_weight),
        )

        # 2nd run
        y_true = (np.random.uniform(0, 1, size=(5, 6, 7)) > 0.5).astype(np.float32)
        y_pred = np.random.uniform(0, 1, size=(5, 6, 7))
        sample_weight = np.random.uniform(0, 1, size=(5, 6))

        assert np.allclose(
            tm(y_true, y_pred, sample_weight=sample_weight),
            em.call_with_defaults()(y_true, y_pred, sample_weight=sample_weight),
        )



from elegy.module import Module
import typing as tp

import haiku as hk
import jax.numpy as jnp
import numpy as np


from elegy import types


def _infer_shape(output_shape, dimensions):
    """
    Replaces the -1 wildcard in the output shape vector.

    This function infers the correct output shape given the input dimensions.

    Args:
        output_shape: Output shape.
        dimensions: List of input non-batch dimensions.

    Returns:
        Tuple of non-batch output dimensions.
    """
    # Size of input.
    n = np.prod(dimensions)
    # Size of output where defined.
    v = np.array(output_shape)
    m = abs(np.prod(v))
    # Replace wildcard.
    v[v == -1] = n // m
    return tuple(v)


class Reshape(Module):
    """
    Reshapes input Tensor, preserving the batch dimension.

    For example, given an input tensor with shape `[B, H, W, C, D]`:

    ```python
    B, H, W, C, D = range(1, 6)
    x = jnp.ones([B, H, W, C, D])
    ```

    The default behavior when `output_shape` is `(-1, D)` is to flatten
    all dimensions between `B` and `D`:

    ```python
    mod = elegy.nn.Reshape(output_shape=(-1, D))
    assert mod(x).shape == (B, H*W*C, D)
    ```

    You can change the number of preserved leading dimensions via
    `preserve_dims`:

    ```python
    mod = elegy.nn.Reshape(output_shape=(-1, D), preserve_dims=2)
    assert mod(x).shape == (B, H, W*C, D)
    mod = elegy.nn.Reshape(output_shape=(-1, D), preserve_dims=3)
    assert mod(x).shape == (B, H, W, C, D)
    mod = elegy.nn.Reshape(output_shape=(-1, D), preserve_dims=4)
    assert mod(x).shape == (B, H, W, C, 1, D)
    ```
    """

    def __init__(self, output_shape: types.Shape, preserve_dims: int = 1, **kwargs):
        """
        Constructs a `Reshape` module.

        Args:
            output_shape: Shape to reshape the input tensor to while preserving its
                first `preserve_dims` dimensions. When the special value -1 appears in
                `output_shape` the corresponding size is automatically inferred. Note
                that -1 can only appear once in `output_shape`.
                To flatten all non-batch dimensions use `Flatten`.
            preserve_dims: Number of leading dimensions that will not be reshaped.
            kwargs: Additional keyword arguments passed to Module.

        Raises:
            ValueError: If `preserve_dims` is not positive.
        """
        super().__init__(**kwargs)
        if preserve_dims <= 0:
            raise ValueError("Argument preserve_dims should be >= 1.")
        if output_shape.count(-1) > 1:
            raise ValueError("-1 can only occur once in `output_shape`.")

        self.output_shape = tuple(output_shape)
        self.preserve_dims = preserve_dims

    def call(self, inputs: np.ndarray) -> np.ndarray:
        """
        Arguments:
            inputs: the array to be reshaped.

        Returns:
            A reshaped array.
        """
        if inputs.ndim <= self.preserve_dims:
            return inputs

        if -1 in self.output_shape:
            reshaped_shape = _infer_shape(
                self.output_shape, inputs.shape[self.preserve_dims :]
            )
        else:
            reshaped_shape = self.output_shape
        shape = inputs.shape[: self.preserve_dims] + reshaped_shape
        return jnp.reshape(inputs, shape)


class Flatten(Reshape):
    """
    Flattens the input, preserving the batch dimension(s).

    By default, Flatten combines all dimensions except the first.
    Additional leading dimensions can be preserved by setting preserve_dims.

    ```python
    x = jnp.ones([3, 2, 4])
    flat = elegy.nn.Flatten()
    assert flat(x).shape == (3, 8)
    ```

    When the input to flatten has fewer than `preserve_dims` dimensions it is
    returned unchanged:

    ```python
    x = jnp.ones([3])
    assert flat(x).shape == (3,)
    ```
    """

    def __init__(
        self,
        preserve_dims: int = 1,
        name: tp.Optional[str] = None,
    ):
        super().__init__(output_shape=(-1,), preserve_dims=preserve_dims, name=name)



from unittest import TestCase

import jax.numpy as jnp
import tensorflow.keras as tfk

import elegy


class BinaryCrossentropyTest(TestCase):
    #
    def test_basic(self):

        # Input:  true (y_true) and predicted (y_pred) tensors
        y_true = jnp.array([[0.0, 1.0], [0.0, 0.0]])
        y_pred = jnp.array([[0.6, 0.4], [0.4, 0.6]])

        # Standard BCE, considering prediction tensor as probabilities
        bce = elegy.metrics.BinaryCrossentropy()
        result = bce.call_with_defaults()(
            y_true=y_true,
            y_pred=y_pred,
        )
        assert jnp.isclose(result, 0.815, rtol=0.01)

        # Standard BCE, considering prediction tensor as logits
        y_logits = jnp.log(y_pred) - jnp.log(1 - y_pred)
        bce = elegy.metrics.BinaryCrossentropy(from_logits=True)
        result_from_logits = bce.call_with_defaults()(y_true, y_logits)
        assert jnp.isclose(result_from_logits, 0.815, rtol=0.01)
        assert jnp.isclose(result_from_logits, result, rtol=0.01)

        # BCE using sample_weight
        bce = elegy.metrics.BinaryCrossentropy()
        result = bce.call_with_defaults()(
            y_true, y_pred, sample_weight=jnp.array([1.0, 0.0])
        )
        assert jnp.isclose(result, 0.916, rtol=0.01)

    #
    def test_compatibility(self):

        # Input:  true (y_true) and predicted (y_pred) tensors
        y_true = jnp.array([[0.0, 1.0], [0.0, 0.0]])
        y_pred = jnp.array([[0.6, 0.4], [0.4, 0.6]])

        # Standard BCE, considering prediction tensor as probabilities
        bce_elegy = elegy.metrics.BinaryCrossentropy()
        bce_tfk = tfk.metrics.BinaryCrossentropy()
        assert jnp.isclose(
            bce_elegy.call_with_defaults()(y_true, y_pred),
            bce_tfk(y_true, y_pred),
            rtol=0.0001,
        )

        # Standard BCE, considering prediction tensor as logits
        y_logits = jnp.log(y_pred) - jnp.log(1 - y_pred)
        bce_elegy = elegy.metrics.BinaryCrossentropy(from_logits=True)
        bce_tfk = tfk.metrics.BinaryCrossentropy(from_logits=True)
        assert jnp.isclose(
            bce_elegy.call_with_defaults()(y_true, y_logits),
            bce_tfk(y_true, y_logits),
            rtol=0.0001,
        )

        # BCE using sample_weight
        bce_elegy = elegy.metrics.BinaryCrossentropy()
        bce_tfk = tfk.metrics.BinaryCrossentropy()
        assert jnp.isclose(
            bce_elegy.call_with_defaults()(
                y_true, y_pred, sample_weight=jnp.array([1, 0])
            ),
            bce_tfk(y_true, y_pred, sample_weight=jnp.array([1, 0])),
            rtol=0.0001,
        )



# Implementation based on tf.keras.engine.data_adapter.py
# https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/keras/engine/data_adapter.py


import typing as tp

import numpy as np

from .array_adapter import ArrayDataAdapter
from .data_adapter import DataAdapter

scalar_types = (float, int, str)


class ListsOfScalarsDataAdapter(DataAdapter):
    """Adapter that handles lists of scalars and lists of lists of scalars."""

    @staticmethod
    def can_handle(x, y=None):
        handles_x = ListsOfScalarsDataAdapter._is_list_of_scalars(x)
        handles_y = True
        if y is not None:
            handles_y = ListsOfScalarsDataAdapter._is_list_of_scalars(y)
        return handles_x and handles_y

    @staticmethod
    def _is_list_of_scalars(inp):
        if isinstance(inp, scalar_types):
            return True
        if isinstance(inp, (list, tuple)):
            return ListsOfScalarsDataAdapter._is_list_of_scalars(inp[0])
        return False

    def __init__(
        self, x, y=None, sample_weights=None, batch_size=None, shuffle=False, **kwargs
    ):
        super(ListsOfScalarsDataAdapter, self).__init__(x, y, **kwargs)
        x = np.asarray(x)
        if y is not None:
            y = np.asarray(y)
        if sample_weights is not None:
            sample_weights = np.asarray(sample_weights)

        self._internal_adapter = ArrayDataAdapter(
            x,
            y=y,
            sample_weights=sample_weights,
            batch_size=batch_size,
            shuffle=shuffle,
            **kwargs
        )

    def get_dataset(self):
        return self._internal_adapter.get_dataset()

    def get_size(self):
        return self._internal_adapter.get_size()

    @property
    def batch_size(self):
        return self._internal_adapter.batch_size

    def has_partial_batch(self):
        return self._internal_adapter.has_partial_batch()

    @property
    def partial_batch_size(self):
        return self._internal_adapter.partial_batch_size

    def should_recreate_iterator(self):
        return True



# Implementation based on tf.keras.callbacks.py
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/callbacks.py
from .callback import Callback


class History(Callback):
    """Callback that records events into a `History` object.

    This callback is automatically applied to
    every Keras model. The `History` object
    gets returned by the `fit` method of models.
    """

    def __init__(self):
        super(History, self).__init__()
        self.history = {}

    def on_train_begin(self, logs=None):
        self.epoch = []

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        self.epoch.append(epoch)
        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)

        # Set the history attribute on the model after the epoch ends. This will
        # make sure that the state which is set is the latest one.
        self.model.history = self



from elegy import types
import typing as tp

import jax.numpy as jnp

from elegy.losses.mean_absolute_error import mean_absolute_error
from elegy.metrics.mean import Mean


class MeanAbsoluteError(Mean):
    """
    Computes the cumulative mean absolute error between `y_true` and `y_pred`.

    Usage:
    ```python
    mae = elegy.metrics.MeanAbsoluteError()

    result = mae(y_true=jnp.array([1, 1, 1, 1]), y_pred=jnp.array([0, 1, 1, 1]))
    assert result == 0.25

    result = mae(y_true=jnp.array([1, 1, 1, 1]), y_pred=jnp.array([1, 0, 0, 0]))
    assert result == 0.5
    ```

    Usage with elegy API:

    ```python
    model = elegy.Model(
        module_fn,
        loss=elegy.losses.CategoricalCrossentropy(),
        metrics=elegy.metrics.MeanAbsoluteError(),
    )
    ```
    """

    def __init__(self, on: tp.Optional[types.IndexLike] = None, **kwargs):
        """
        Creates a `MeanAbsoluteError` instance.

        Arguments:
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(on=on, **kwargs)

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """
        Accumulates metric statistics. `y_true` and `y_pred` should have the same shape.

        Arguments:
            y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.
            y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.
            sample_weight: Optional `sample_weight` acts as a
                coefficient for the metric. If a scalar is provided, then the metric is
                simply scaled by the given value. If `sample_weight` is a tensor of size
                `[batch_size]`, then the metric for each sample of the batch is rescaled
                by the corresponding element in the `sample_weight` vector. If the shape
                of `sample_weight` is `[batch_size, d0, .. dN-1]` (or can be broadcasted
                to this shape), then each metric element of `y_pred` is scaled by the
                corresponding value of `sample_weight`. (Note on `dN-1`: all metric
                functions reduce by 1 dimension, usually the last axis (-1)).
        Returns:
            Array with the cumulative accuracy.
        """

        return super().call(values=mean_absolute_error(y_true=y_true, y_pred=y_pred))



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Moving averages."""

import re

import jax
import jax.numpy as jnp
from haiku._src import data_structures
import numpy as np

from elegy import initializers, module, types


class ExponentialMovingAverage(module.Module):
    """Maintains an exponential moving average.

    This uses the Adam debiasing procedure.
    See https://arxiv.org/pdf/1412.6980.pdf for details.
    """

    counter: np.ndarray
    hidden: np.ndarray
    average: np.ndarray

    def __init__(self, decay, zero_debias=True, warmup_length=0, **kwargs):
        """Initializes an ExponentialMovingAverage module.

        Args:
            decay: The chosen decay. Must in [0, 1). Values close to 1 result in slow
                decay; values close to 0 result in fast decay.
            zero_debias: Whether to run with zero-debiasing.
            warmup_length: A positive integer, EMA has no effect until
                the internal counter has reached `warmup_length` at which point the
                initial value for the decaying average is initialized to the input value
                after `warmup_length` iterations.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(**kwargs)
        self._decay = decay
        if warmup_length < 0:
            raise ValueError(
                f"`warmup_length` is {warmup_length}, but should be non-negative."
            )
        self._warmup_length = warmup_length
        self._zero_debias = zero_debias
        if warmup_length and zero_debias:
            raise ValueError(
                "Zero debiasing does not make sense when warming up the value of the "
                "average to an initial value. Set zero_debias=False if setting "
                "warmup_length to a non-zero value."
            )

    def _cond(self, cond, t, f, dtype):
        """Internal, implements jax.lax.cond without control flow."""
        c = cond.astype(dtype)
        return c * t + (1.0 - c) * f

    def initialize(self, value):
        """If uninitialized sets the average to ``zeros_like`` the given value."""
        self.add_parameter(
            "hidden", lambda: jnp.zeros(value.shape, jnp.float32), trainable=False
        )
        self.add_parameter(
            "average", lambda: jnp.zeros(value.shape, jnp.float32), trainable=False
        )

    def call(self, value: jnp.ndarray, update_stats=True):
        """Updates the EMA and returns the new value.

        Args:
          value: The array-like object for which you would like to perform an
            exponential decay on.
          update_stats: A Boolean, whether to update the internal state
            of this object to reflect the input value. When `update_stats` is False
            the internal stats will remain unchanged.

        Returns:
          The exponentially weighted average of the input value.
        """
        if not isinstance(value, jnp.ndarray):
            value = jnp.asarray(value)

        counter = self.add_parameter(
            "counter",
            lambda: jnp.broadcast_to(-self._warmup_length, ()).astype(jnp.int32),
            trainable=False,
        )
        counter += 1

        decay = jax.lax.convert_element_type(self._decay, value.dtype)
        if self._warmup_length > 0:
            decay = self._cond(counter <= 0, 0.0, decay, value.dtype)

        one = jnp.ones([], value.dtype)
        hidden = self.add_parameter(
            "hidden",
            lambda: jnp.zeros(value.shape, jnp.float32),
            trainable=False,
        )
        hidden = hidden * decay + value * (one - decay)

        average = hidden
        if self._zero_debias:
            average /= one - jnp.power(decay, counter)

        self.add_or_update_parameter("average", average, trainable=False)

        if update_stats:
            self.update_parameter("counter", counter)
            self.update_parameter("hidden", hidden)
            self.update_parameter("average", average)

        return average


class EMAParamsTree(module.Module):
    """Maintains an exponential moving average for all parameters in a tree.

    While ExponentialMovingAverage is meant to be applied to single parameters
    within a function, this class is meant to be applied to the entire tree of
    parameters for a function.

    Given a set of parameters for some network:

    ```python
    >>> import elegy
    >>> x = jnp.ones([1, 1])
    >>> linear = elegy.nn.Linear(10)
    >>> y, params = linear.init(rng=elegy.RNGSeq(42))(x)

    ```

    You might use the EMAParamsTree like follows:

    ```python
    >>> ema = elegy.nn.EMAParamsTree(0.2)
    >>> _, ema_state = ema.init()(params)
    >>> ema_params, ema_state = ema.apply(ema_state)(params)

    ```

    Here, we are transforming a Haiku function and constructing its parameters via
    an init_fn as normal, but are creating a second transformed function which
    expects a tree of parameters as input. This function is then called with
    the current parameters as input, which then returns an identical tree with
    every parameter replaced with its exponentially decayed average. This
    ema_params object can then be passed into the `network_fn` as usual, and will
    cause it to run with EMA weights.
    """

    def __init__(
        self, decay, zero_debias=True, warmup_length=0, ignore_regex="", **kwargs
    ):
        """Initializes an EMAParamsTree module.

        Args:
            decay: The chosen decay. Must in [0, 1). Values close to 1 result in slow
                decay; values close to 0 result in fast decay.
            zero_debias: Whether to run with zero-debiasing.
            warmup_length: A positive integer, EMA has no effect until
                the internal counter has reached `warmup_length` at which point the
                initial value for the decaying average is initialized to the input value
                after `warmup_length` iterations.
            ignore_regex: A string. Any parameter in the tree whose name matches this
                regex will not have any moving average applied to it. The empty string
                means this module will EMA all parameters.
            kwargs: Additional keyword arguments passed to Module.
        """
        super().__init__(**kwargs)
        self._decay = decay
        self._zero_debias = zero_debias
        self._warmup_length = warmup_length
        self._ignore_regex = ignore_regex

    def call(self, tree, update_stats=True):
        def maybe_ema(k, v):
            if self._ignore_regex and re.match(self._ignore_regex, k):
                return v
            else:
                ema_name = k.replace("/", "__").replace("~", "_tilde_")
                return ExponentialMovingAverage(
                    self._decay, self._zero_debias, self._warmup_length, name=ema_name
                )(v, update_stats=update_stats)

        # We want to potentially replace params with EMA'd versions.
        new_values = {}
        for module_name, param_dict in tree.items():
            new_values[module_name] = {
                k: maybe_ema("/".join([module_name, k]), v)
                for k, v in param_dict.items()
            }
        return data_structures.to_immutable_dict(new_values)
